#include <ctype.h>
#include <stdio.h>
#include <math.h>
#include <pgm.h>
#include "bitcodes.h"

typedef unsigned int UInt32;

#define MAX_SQUARES 5000
#define MAX_LINE_PIXEL  17000   // a line all the way thru symbol
#define MAX_MOD_PIXEL   1700   // a line thru one modules worth = 17 bars
#define MAX_X_PIXEL 17000
#define MAX_Y_PIXEL 5000
#define OUT_MAX_X_PIXEL 6000
#define OUT_MAX_Y_PIXEL 2000
#define MAX_XPIXEL_MODULE  900    // number of x pixels in a 17 wide module
#define MAX_YPIXEL_MODULE 300    // number of y pixels in a 17 wide module
#define MAX_SYMBOL_ROWS  100  // standard limits it to 90
#define MAX_SYMBOL_COLS  330  // min 3 rows 929/3 = 310 
#define MAX_LINES_X 300
#define MAX_LINES_Y 300
#define MAX_TOT_LR   5000
#define MAX_BARS  2000
#define MAX_PB 1500
#define MAX_MOD_BARS 5     // at most 5 bars in a module ( actually 4 )
#define MAX_LINE_SCANS 120
#define TRUE 1
#define FALSE 0

// scan defines

 #define MAX_BYTE_INDEX 1024
 #define MAX_CODE 2500

 #define ALPHA 0
 #define LOWER 1
 #define MIXED 2
 #define PUNCT 3
 #define PUNCT_SH 4

 #define MAX_ROW 91
 #define MAX_COL 512
 #define NULLCHAR '\0'

 #define MAX_RAW_CODES 1500
 #define MIN_ROW_HEIGHT  4
 #define MIN_COL_WIDTH   4
 #define EDGE_THRESHOLD  10

 #define MAX_PIXEL_Y 2048
 #define MAX_PIXEL_X 2048


 #define NN 1024
 #define PRIM 1
 #define GPRIME 929
 #define KK 32
 #define A0 928
 #define Ldec 1

#define sgn(x) ((x<0)?-1:((x>0)?1:0)) /* macro to return the sign of a
                                         number */
void  write_pgm_outpix_final( int num_cols, int num_rows);
void write_pgm_outmod(char *namestrin,  int num_cols, int num_rows);
void write_pgm_outcol(char *namestrin,  int num_cols, int num_rows);
void write_pgm_out( int num_cols, int num_rows);
void write_pgm_outpix( int num_cols, int num_rows);

FILE *wr_data_file;
FILE *infile1, *outfile1,*infile;

int fail;
char inargstr[120];

char pgm_infile_str[120];

int write_mods;      // flags to control writing of pgm outputs
int write_final_pix;  // write the symbol after unslanting 
int write_leftcol;    // write the left column of the symbol
int write_cols;       // write the individual columns
int write_outpix;        // write the symbol before unslanting 
int write_data;         // write the bitcodes for each row, and col

int box_x_incr;   // size of box used to find the symbol
int box_y_incr;   // inside the white space that should surround it

int slope_array[100];
int fig_flag;

int tot_lr[MAX_TOT_LR];      // used by get_columns_start
int tot_lr_in[MAX_TOT_LR];    // 1st derivative of tot_lr

int begin_left_col;   // beginning pixel of left column
int begin_right_bars;  // beginning pixel of right bars

int min_box_x;
int min_box_y;
int max_box_x;
int max_box_y;

int bytes_per_pix;      // bytes per pixel

int col_begin_x[400];     // beginning x value of each column

float onewidthf;
int onewidth;
float onewf;
int num_sym_cols;

int anal_mod_index;     // successful index for analyze module
int rule_mod_index;     // successful index for rule_checker
int slide_mod_index;
int xline_mod_index;

int symbol_data[MAX_SYMBOL_ROWS][MAX_SYMBOL_COLS]; 
             //  put the codes for rows and columns here

int bestfig;
int pixel_out_xlen;   // size of the pix_out in x direction
int pixel_out_ylen;   // size of the pix_out in y direction

int dir;
int slope;
int pix_slope;       // final correction slope
float pix_slopef;

int best_normal_x;
int best_normal_y;
int best_normal_dir;
int best_onewidth;
int best_begin_left_col;
int best_begin_right_bars;
int best_num_sym_cols;
int best_pix_slope;
int best_normal_slope;
int best_pixel_out_xlen;
int best_pixel_out_ylen;
float best_onewf;

int left_highy;     // highest y value in pix_out at middle of left fat bar
int right_highy;    // highest y value in pix_out at middle of right fat bar
int left_highx;
int right_highx;
int y_begin;       // this is the estimated center of the top row

int scan_upleft_x;   // these are the beginning points of places to scan from
int scan_upleft_y;

int scan_cenleft_x;
int scan_cenleft_y;

int scan_lowleft_x;
int scan_lowleft_y;
int out_cnt;

int save_line_cnt;  // should not be more than 36

int dark_sqindex;      // indexs all the dark squares in the original image
int dark_threshold;    // average grey value of a dark square
int line_threshold;    
int lpixel;
int pixel_cnt;    
int post_pixel_cnt;
int mod_pixel_cnt;

    
int symbol_height;    // pixel count height of symbol at left wide bar

int number_of_rows;   // number of rows in pixel map
int number_of_cols;  // number of columns in pixel map
int row_height_avg;   // avg of row_height high and low estimate
int best_row_height;  // best estimate of row height
int best_col_width_avg;  // best estimate of avg column width

int pix_col_yline[MAX_LINES_Y];  // y co-ordinant of lines in column data
int pix_col_xline[MAX_LINES_X];  // x co-ordinant of lines in column data
int pix_col_ydiffs[MAX_LINES_Y];
int vert_line_cnt;

void draw_normal(int nx, int ny, int nslope, int ndir, int ylen);

void line_fast_norm(int x1, int y1, int x2, int y2);

void find_height(int nx, int ny, int nslope, int ndir, int ylen);

void final_pix(int xlen, int ylen );

void find_left_edge(int nx, int ny, int nslope, int ndir, int ylen, int xlen);

void get_a_column(int indexin, int startx, int onewid);

struct pixelod
{
  int greyval;
} pix_in[MAX_X_PIXEL * MAX_Y_PIXEL];

struct pixeluu
{
  int greyval;
} pix_module[MAX_XPIXEL_MODULE][MAX_YPIXEL_MODULE];

struct pixeltt
{
  int greyval;
} pix_cols[OUT_MAX_X_PIXEL][OUT_MAX_Y_PIXEL];

struct pixelzz
{
  int greyval;
} pix_out[OUT_MAX_X_PIXEL][OUT_MAX_Y_PIXEL];

struct pixelqq
{
  int greyval;
} pix_final[OUT_MAX_X_PIXEL][OUT_MAX_Y_PIXEL];
 
struct pixeldo 
{
  int greyval;
} pixel_array[MAX_X_PIXEL][MAX_Y_PIXEL];

struct darksqstuff
{
  int xloc;
  int yloc;
} dark_squares[MAX_SQUARES];

// codes from bit analysis go here

struct pb_stuff
{
  int rowval;
  int colval;
  int codeval;
} pb_array[MAX_PB];

int bar_cnt;   // count of bars
int pb_index;
int mod_bar_cnt;   // count of bars in a particular module

int b_box_xlen;
int b_box_ylen;
int b_box_maxx;
int b_box_maxy;
int b_box_minx;
int b_box_miny;
float max_theta;	/* Maximum theta range */
int barwidth;

struct barstuff
{
  int beginx;
  int beginy;
  int endx;
  int endy;
  int space_width;  // width of space before bar
  int width;
  int lpixel; 
}  bar_array[MAX_BARS];

struct mbarstuff
{
  int beginx;
  int beginy;
  int endx;
  int endy;
  int space_width;  // width of space before bar
  int width;
  int lpixel; 
}  mod_bar_array[MAX_MOD_BARS];

 int mod_bit_array[40];

struct linesavestuff
{
  int pixelcnt;
  int slope;
  int dir;
  int beginx;
  int beginy;
  int has_8111;   // begin 
  int has_7111;   // end
  int endlpixel;
  int endwidth;    // width of 7 wide end bar
  int beginwidth;   // width of begin 8 wide bar
  int barcnt;
  int bars_found;
  int normal_val;          // find the number of pixels hit by normal
  int line_vals[MAX_LINE_PIXEL];
  int line_x[MAX_LINE_PIXEL];
  int line_y[MAX_LINE_PIXEL];
  struct barstuff tbar[8];
  int fig;
} save_line_data[MAX_LINE_SCANS];

gray **greydata;

int norm_fdark_ind;   // number of pixels before first dark in norm
                      // line between one of the left start bars

int norm_line_data[MAX_LINE_PIXEL];
int normal_x[MAX_LINE_PIXEL];
int normal_y[MAX_LINE_PIXEL];

int post_line_data[MAX_LINE_PIXEL];
int post_x[MAX_LINE_PIXEL];
int post_y[MAX_LINE_PIXEL];

int line_data[MAX_LINE_PIXEL];
int mod_line_data[MAX_MOD_PIXEL];
int temp_data[MAX_LINE_PIXEL];

int line_x[MAX_LINE_PIXEL];
int line_y[MAX_LINE_PIXEL];

int mod_line_x[MAX_MOD_PIXEL];

int est_bar_height;   // estimate of bar height found from diffy
int bar_height;       // best estimate bar height found from hieght of symbol
                      // divided by number rows

// scan global variables

 int pixel_dark( int inx, int iny);
 int Alpha_to[1024];
 int Index_of[1024];
 int powers_of_3[1024];
 int log_of_3[1024];
 int verbose_out;
 int use_image;   // set true if getting bit image from file

 typedef struct raw_code_type {
    int code;
    int row;
    int col;
  } raw_codes_type;

 raw_codes_type raw_codes[MAX_RAW_CODES];
 int total_codes;
 int top_left_corn_x;
 int top_left_corn_y;

 int sum_array[20000];
 int x_width;            // pixel width of scan in x direction
 int y_height;          // pixel height of scan in y direction
 int row_last_code;     // last code in row
 int col_count;
 int pixel_x;
 int pixel_y;
 int expected_cols;
 int col_width;
 float fcol_width;
 int row_height;
 int barcode_height;
 int barcode_width;
 int code_index;
 int code_index_t;      // code_index using t array
 int scan_code_array[128][40];
 int top_data_corner;
 int left_data_corner;
 int left_edge;
 int right_edge;
 int top_edge;
 int bottom_edge;
 int end_bar_found;
 int x_center;
 int y_center;
 int row_rem;
 int synd_array[512];
 int rows_cols[MAX_ROW][MAX_COL];
 int number_of_codes;
 int rows_from_raw;
 int columns_from_raw;
 int byte_compress_mode;
 int text_compress_mode;
 int numeric_compress_mode;
 int code_array[MAX_CODE];
 int code_array_size;
 int byte_array[MAX_BYTE_INDEX];
 struct b913_stuff
  { 
   char str913[20];
  }  byte_913_str[MAX_BYTE_INDEX];

 char punct_chars[32];
 char mixed_chars[32];
 int codes_bc[20];
 int codes_nc[20];
 int codeindex;
 int code;
 int number_check_words;
 int end_found;
 int raw_row_col [90] [100];
 char powers_of_900[16][50];   
 char powers_of_900h[10][50];
 char powers_of_900h_x10[10][50];
 char powers_of_900h_x100[10][50];
 char result[256];
 int data[1024];
 int erase_loc[MAX_RAW_CODES];  // where are the erasures?
 int erase_count;

 

void init_slopes()
{
  slope_array[0] = 0;
  slope_array[1] = 2;
  slope_array[2] = 4;
  slope_array[3] = 6;
  slope_array[4] = 10;
  slope_array[5] = 12;
  slope_array[6] = 14;
  slope_array[7] = 16;
  slope_array[8] = 18;
  slope_array[9] = 20;
  slope_array[10] = 23 ;
  slope_array[11] = 26;
  slope_array[12] = 29;
  slope_array[13] = 32;
  slope_array[14] = 36;
  slope_array[15] = 40;
  slope_array[16] = 44;
  slope_array[17] = 48;
  slope_array[19] = 52;
  slope_array[20] = 56;
  // slope_array[12] = 214;
  // slope_array[13] = 274;
  // slope_array[14] = 373;
}

// find all the squares in the image that are darker than a certain
//  thresshold.  Put the center x,y location of these squares in the 
//  array dark_squares

void find_dark_squares(int x_left, int y_top, int x_size, int y_size)
{

  int ii;
  int jj;
  int debug;
 
  int x_center;
  int y_center;
  int y_tot;
  int xy_tot;
  int y_avg;
  int xy_avg;

  debug = 0;  
  if ( debug )
    {
     printf("In find dark sq , in x,y = %d %d \n",  x_left, y_top);
    }

  xy_tot = 0;
  for(ii = x_left; ii < x_left + x_size; ii += 1)
    {

      y_tot = 0;
      for ( jj = y_top; jj > y_top - y_size; jj -=1 )
        {
          y_tot = y_tot +  pixel_array[ii][jj].greyval;
     
        }
       if ( y_size > 0.0)
	{
         y_avg = y_tot / y_size;
        }
       else
	 {
	   printf("\n y_size not positive = %f \n", y_size);
           exit(0);
         }
       xy_tot = xy_tot + y_avg;
    }

  if ( x_size > 0 )
    {
     xy_avg = xy_tot / x_size;
    }
  else
    {
      printf("\n x_size not positive = %f \n", x_size);
      exit(0);
    }
  

  if ( debug) { printf("xy_avg = %d \n", xy_avg); }

  if ( xy_avg > dark_threshold)
    {
      x_center = x_left + (( x_size) / 2);
      y_center = y_top - ((y_size) / 2);

      if ( dark_sqindex < MAX_SQUARES)
	{
         dark_squares[dark_sqindex].xloc = x_center;
         dark_squares[dark_sqindex].yloc = y_center;
         dark_sqindex += 1;
         if (debug)
	   {
            printf("Found a darks square at = %d %d \n", x_center, y_center );
           }
        }
      else
        {
	  printf("Error: Number of dark boxes too big = %d \n",
		 dark_sqindex);
        }
    }
}

// draw horizontal and vertical lines
//

void draw_pixel_line(int width, int beginx, int beginy, int endx, int endy )
{
  int ii,jj;


  if (( beginx - endx )  == 0)  // vertical
    {
      if ( endy > beginy )          
        {
	  for ( ii = beginy; ii < endy; ii += 1)
	    {
	      for ( jj = 0; jj < width; jj += 1)
                {
		  if ( ii + jj < MAX_Y_PIXEL)
		    {
		     pixel_array[beginx][ii+jj].greyval = 510;
                    }
                }
            } 
        }
      else
        {
          for ( ii = endy; ii < beginy; ii += 1)
	    {
	      for ( jj = 0; jj < width; jj += 1)
                {
                  if ( ii + jj < MAX_Y_PIXEL)
		    {
		     pixel_array[beginx][ii+jj].greyval = 510;
                    }
                }
            } 
        }
    }


  if (( beginy - endy )  == 0)  // horizontal
    {
      if ( endx > beginx )          
        {
	  for ( ii = beginx; ii < endx; ii += 1)
	    {
	      for ( jj = 0; jj < width; jj += 1)
                {
                  if ( ii + jj < MAX_X_PIXEL)
		    {
		      pixel_array[ii + jj ][beginy].greyval = 510;
                    }
                }
            } 
        }
      else
        {
          for ( ii = endx; ii < beginx; ii += 1)
	    {
	      for ( jj = 0; jj < width; jj += 1)
                {
                  if ( ii + jj < MAX_X_PIXEL)
		    {
		     pixel_array[ii + jj][beginy].greyval = 510;
                    }
                }
            } 
        }
    }


}

//
//  Search thru all the dark squares found and find the 
//      min and max of the x and y directions to make a box

void find_dark_sq_box()

{
  int ii;
  int debug;

  min_box_x = MAX_X_PIXEL;
  min_box_y = MAX_Y_PIXEL;

  max_box_x = 0;
  max_box_y = 0;

  debug = 0;
  
  for (ii = 0; ii < dark_sqindex; ii += 1)
    {
      if ( dark_squares[ii].xloc > max_box_x)
        {
          max_box_x = dark_squares[ii].xloc;
	}
      if ( dark_squares[ii].yloc > max_box_y)
        {
          max_box_y = dark_squares[ii].yloc;
	}
      if ( dark_squares[ii].xloc < min_box_x)
        {
          min_box_x = dark_squares[ii].xloc;
	}
      if ( dark_squares[ii].yloc < max_box_y)
        {
          min_box_y = dark_squares[ii].yloc;
	}

    }

  min_box_x = min_box_x - box_x_incr;
  min_box_y = min_box_y - box_y_incr;
  max_box_x = max_box_x + box_x_incr;
  max_box_y = max_box_y + box_y_incr;

  if  (debug)
    {
     printf("In find dark square bounding box \n");
     printf(" min x,y = %d %d \n", min_box_x, min_box_y);
     printf(" max x,y = %d %d \n", max_box_x, max_box_y);
    }

  // draw a box around the dark squares and put this in pixel_array
  // then you can dump out the pixel_array and see how good it is

  if ( debug)
    {
     draw_pixel_line( 4, max_box_x, max_box_y, min_box_x, max_box_y );
     draw_pixel_line( 4, max_box_x, max_box_y, max_box_x, min_box_y );
     draw_pixel_line( 4, min_box_x, min_box_y, max_box_x, min_box_y );
     draw_pixel_line( 4, min_box_x, min_box_y, min_box_x, max_box_y );
    }

}  // find dark square box

// expand the box around the symbol by 25 %
//   may not be necessary...

void expand_dark_box25()
{
  int x_length;
  int y_length;
  
  int new_x_len;
  int new_y_len;
  
  int new_min_x;
  int new_min_y;
  int new_max_x;
  int new_max_y;


  x_length = max_box_x - min_box_x;
  
  if ( x_length < ( MAX_X_PIXEL / 50) )
    {
      printf("Symbol too small or missing \n");
    }

  y_length = max_box_y - min_box_y;
  
  if ( y_length < ( MAX_Y_PIXEL / 50) )
    {
      printf("Symbol too small or missing \n");
    }
  
  // expand by 25%

  new_x_len = x_length + (x_length / 4);
  new_y_len = y_length + (y_length / 4);


  new_min_x =  max_box_x - ( x_length / 2 ) - new_x_len;
  new_min_y =  max_box_y - ( y_length / 2 ) - new_y_len;

  if ( new_min_x > 0)
    {
      new_max_x =  new_min_x + new_x_len;
    }

  if ( new_min_y > 0)
    {
      new_max_y =  new_min_y + new_y_len;
    }


  if ( new_min_x < 0)
    {
      // could try to expand by 10 % here
      new_min_x = min_box_x;
    }
  
  if ( new_min_y < 0)
    {
      // could try to expand by 10 % here
      new_min_y = min_box_y;
    }

  if ( new_max_x > MAX_X_PIXEL)
    {
      // could try to expand by 10 % here
      new_max_x = max_box_x;
    }

  
  if ( new_min_y > MAX_Y_PIXEL)
    {
      // could try to expand by 10 % here
      new_max_y = max_box_y;
    }

  b_box_maxx = new_max_x;
  b_box_maxy = new_max_y;  

  b_box_minx = new_min_x;
  b_box_miny = new_min_y;  
  
  b_box_xlen = ( b_box_maxx - b_box_minx);
  b_box_ylen = ( b_box_maxy - b_box_miny);
}

void get_box_len()
{
  int x_length;
  int y_length;
  
  int new_x_len;
  int new_y_len;
  
  int new_min_x;
  int new_min_y;
  int new_max_x;
  int new_max_y;
  int debug;


  debug = 0;

  x_length = max_box_x - min_box_x;
  
  if ( x_length < ( MAX_X_PIXEL / 50) )
    {
      printf("Symbol too small or missing \n");
    }

  y_length = max_box_y - min_box_y;
  
  if ( y_length < ( MAX_Y_PIXEL / 50) )
    {
      printf("Symbol too small or missing \n");
    }
  


  new_x_len = x_length ;
  new_y_len = y_length ;


  new_min_x = min_box_x;
  new_min_y = min_box_y;
  new_max_x = max_box_x;
  new_max_y = max_box_y;

  b_box_maxx = new_max_x;
  b_box_maxy = new_max_y;  

  b_box_minx = new_min_x - (x_length / 20);
  if ( b_box_minx < (min_box_x / 2))
    {
      b_box_minx = (min_box_x / 2);
    }

  b_box_miny = new_min_y;  
  
  b_box_xlen = ( b_box_maxx - b_box_minx);
  b_box_ylen = ( b_box_maxy - b_box_miny);
  if (debug)
    {
      printf("End get_box_length \n");
    }
}

// get 3 points on the left edge of the bounding box
// one at upper left
// one at center
// one at bottom left
// use these as starting points for the line scan
//

void get_scan_points()
{
    
  int debug;
  debug = 0;

  
  scan_upleft_x = b_box_minx;
  scan_upleft_y = b_box_maxy;

  scan_cenleft_x = b_box_minx;
  scan_cenleft_y = b_box_maxy - (b_box_ylen/2);

  scan_lowleft_x = b_box_minx;
  scan_lowleft_y = b_box_miny;

  if (debug) { printf(" In scan points upper left = %d %d \n",
		      scan_upleft_x, scan_upleft_y);
               printf(" In scan points center left = %d %d \n",
		      scan_cenleft_x, scan_cenleft_y);
               printf(" In scan points lower left = %d %d \n",
		      scan_lowleft_x, scan_lowleft_y);
  } 
}


void get_pixel(int inx, int iny)
{
 
  if ( pixel_cnt < MAX_LINE_PIXEL)
    {
     line_data[pixel_cnt] = pixel_array[inx][iny].greyval;
     line_x[pixel_cnt]= inx;
     line_y[pixel_cnt]= iny;

     pixel_cnt += 1;
    }
  else
    {
      printf("Too many pixels in this line \n");
      
    }
}

// get pixel used when taking the normal to the line
//
void get_pixel_norm(int inx, int iny)
{
  int debug;
  
  debug = 0;
 
  if ( pixel_cnt < MAX_LINE_PIXEL)
    {
     norm_line_data[pixel_cnt] = pixel_array[inx][iny].greyval;
  
     normal_x[pixel_cnt] = inx;
     normal_y[pixel_cnt] = iny;
     if (debug)
       {
        if ( pixel_cnt == 0)
         {
	  if (debug ) {
            printf("at 0 normal_x , normal_y = %d %d \n", normal_x[0],
		normal_y[0]);
	  }
         }
       }
     pixel_cnt += 1;
    }
  
}

// get pixel used when taking a line from the pix_out data
//   for final processing
//
void get_pixel_post(int inx, int iny)
{
  int debug;

  debug = 0;

  if ( post_pixel_cnt < MAX_LINE_PIXEL)
    {
      if ( ( inx > 0 ) && ( iny > 0))
	{
          if ( ( inx < pixel_out_xlen ) && ( iny < pixel_out_ylen))
            {
              post_line_data[post_pixel_cnt] = pix_out[inx][iny].greyval;
  
              post_x[post_pixel_cnt] = inx;
              post_y[post_pixel_cnt] = iny;
            }
          else
	    {
              post_line_data[post_pixel_cnt] = 0;
            }
        }
      else
	{
          post_line_data[post_pixel_cnt] = 0;
          post_x[post_pixel_cnt] = -1;
          post_y[post_pixel_cnt] = -1;
        }
     if (debug)
       {
        if ( post_pixel_cnt == 0)
         {
	   printf("at 0 post_x , post_y = %d %d \n", post_x[0],
		post_y[0]);
         }
       }
      post_pixel_cnt += 1;
    }
  
}

void save_line()
{
  int jj;

  if ( save_line_cnt < MAX_LINE_SCANS )
    {
      save_line_data[save_line_cnt].pixelcnt = pixel_cnt;
      save_line_data[save_line_cnt].has_8111 = FALSE;
      save_line_data[save_line_cnt].has_7111 = FALSE;
      save_line_data[save_line_cnt].slope = slope;
      save_line_data[save_line_cnt].dir = dir;
      
      for ( jj= 0; jj < pixel_cnt; jj += 1)
	{
	  save_line_data[save_line_cnt].line_vals[jj]= line_data[jj];
	  //   printf("line_data,jj=%d = %d \n",jj,line_data[jj]);
          save_line_data[save_line_cnt].line_x[jj] = line_x[jj];
          save_line_data[save_line_cnt].line_y[jj] = line_y[jj];
        }
      
    }
  else
    {
      printf("Internal error: Too many left probe lines  = %d\n", save_line_cnt);
      printf("Allowed max = %d \n",MAX_LINE_SCANS);
    }

}

void skip_underthress()
{

  while(( line_data[lpixel] < line_threshold ) && ( lpixel < pixel_cnt))
    {
      lpixel += 1;
      barwidth += 1;
    }
}


// called by analyzer module , used to get the bar
//   widths in the module  using the best xlines in
//   the column

int get_mod_xline(int yval, int rowin , int colin)
{
  int good_space;
  int kk;
  int debug;
  int mid_point;
  int bit_int;
  int space_min;
  int space_max;
  int bit[20];
  int this_mod_index;

  debug = 0;
  good_space = TRUE;
  space_min = (best_onewidth * 3 ) / 4;
  space_max = (( int)( best_onewf * 6)) / 4;

  bit[0] = 0;
  for ( kk = 1; kk < vert_line_cnt; kk += 1)
    {
      bit[kk] = 0;
      if ( (pix_col_xline[kk] - pix_col_xline[kk-1]) < space_min)
	{
	  good_space = FALSE;
          if (debug) { printf("Good space failed : too narrow \n"); }
        }
      if ( (pix_col_xline[kk] - pix_col_xline[kk-1]) > space_max)
	{
	  good_space = FALSE;
          if (debug) { printf("Good space failed : too wide\n"); }
        }
      if ( debug)
        {
          printf("kk = %d pix_col_xline[kk] = %d \n", kk, pix_col_xline[kk]);
        }
    }

  xline_mod_index = -1;
  this_mod_index = -1;

  if (( vert_line_cnt > 15) && ( good_space == TRUE))
    {
     for ( kk = 0; kk < vert_line_cnt ; kk += 1)
      {
       if ( kk > 0)
	{
          mid_point = ( pix_col_xline[kk] + pix_col_xline[kk-1] ) / 2;
        }
       else
        {
          mid_point = pix_col_xline[kk] / 2;
        }
        
        bit[kk] = test_black( mid_point, yval ,1);
        if (debug)
	  {
	    printf("mid_point = %d yval = %d  kk = %d \n",
		   mid_point, yval, kk );
          }
      }
     
   
      bit_int = bits_to_symbol( bit );

     if (debug) { printf("xline bit_int = %x \n", bit_int); }

     this_mod_index = find_index(bit_int,2, rowin, colin);

     if ( this_mod_index != -1 )
       {
         if (debug) { printf("This mod index = %d \n", this_mod_index); }

         xline_mod_index = this_mod_index;         
         return(0);
         }
     else
       {
	 return(-1);
       }
    }
  else
    {
      return(-1);
    }

}


// called by analyzer module , used to get the bar
//   widths in the module

void get_mod_bar(int in_row, int in_col )
{

  int mbarwidth;
  int debug;

  debug = 0;
  if ( debug) { printf("In get_mod_bar \n"); }


  mbarwidth = 0;

  if ( mod_bar_cnt < MAX_MOD_BARS)
    {
      mod_bar_array[mod_bar_cnt].beginx = mod_line_x[lpixel];
      mod_bar_array[mod_bar_cnt].lpixel = lpixel;
    }

  while(( mod_line_data[lpixel] > line_threshold ) && ( lpixel < mod_pixel_cnt))
    {
      lpixel += 1;
      mbarwidth += 1;
    }

  if ( mbarwidth > 2)
    {
      if ( mod_bar_cnt < MAX_MOD_BARS)
	{
	  mod_bar_array[mod_bar_cnt].width = mbarwidth;
          mod_bar_array[mod_bar_cnt].endx = mod_line_x[lpixel];
          
          if (mod_bar_cnt > 0)
            {
             mod_bar_array[mod_bar_cnt].space_width = 
                             (mod_bar_array[mod_bar_cnt].lpixel
		                 - mod_bar_array[mod_bar_cnt-1].lpixel) - 
		mod_bar_array[mod_bar_cnt-1].width;
             
            }
          else
            {
	      mod_bar_array[mod_bar_cnt].space_width = lpixel;
            }
          mod_bar_cnt += 1;
        }
      else
	{
        printf("Warning: More than 5 bars found  in module at row = %d col = %d \n",
	       in_row, in_col ); 
        
        }
    }

  if (debug) { printf("mod_pixel_cnt = %d \n", mod_pixel_cnt); }

  if ( (mod_bar_cnt +1) < MAX_MOD_BARS)
    {
      mod_bar_array[mod_bar_cnt+1].space_width = mod_pixel_cnt -
                                        mod_bar_array[mod_bar_cnt].lpixel
                                      - mod_bar_array[mod_bar_cnt].width;
    }

}

void get_bar()
{

  int gbarwidth;

  gbarwidth = 0;

  if ( bar_cnt < MAX_BARS)
    {
      bar_array[bar_cnt].beginx = line_x[lpixel];
      bar_array[bar_cnt].beginy = line_y[lpixel];
      bar_array[bar_cnt].lpixel = lpixel;
    }
  else
    {
      printf("Too many bars across symbol \n");
      exit(-1);
    }

  while(( line_data[lpixel] > line_threshold ) && ( lpixel < pixel_cnt))
    {
      lpixel += 1;
      gbarwidth += 1;
    }

  if ( gbarwidth > 2)
    {
      if ( bar_cnt < MAX_BARS)
	{
	  bar_array[bar_cnt].width = gbarwidth;
          bar_array[bar_cnt].endx = line_x[lpixel];
          bar_array[bar_cnt].endy = line_y[lpixel];
          if (bar_cnt > 0)
            {
              bar_array[bar_cnt].space_width = (bar_array[bar_cnt].lpixel
		                 - bar_array[bar_cnt-1].lpixel) - 
		bar_array[bar_cnt-1].width;
            }
          else
            {
	      bar_array[bar_cnt].space_width = lpixel;
            }
          bar_cnt += 1;
        }
      else
	{
          printf("Too many bars across symbol\n");
          exit(-1);
        }

    }

}

// look thru the out pix_out data and 
// and find out where the left wide bar ends in the y direction

void find_left_highy(int iny)
{
  int leftx;
  int lefty;
  int debug;

  debug = 0;

  leftx = save_line_data[save_line_cnt].beginwidth / 2;

  lefty = iny;

  if (debug)
    {
      printf("In find left highy , leftx, lefty = %d %d \n",
	     leftx, lefty );
    }

  if (debug) { printf("In find left highy, x,y = %d %d \n", leftx, lefty); }

  while(( lefty > 0) && (( pix_out[leftx][lefty].greyval > dark_threshold) ||
	( pix_out[leftx][lefty-1].greyval > dark_threshold)))
      { 
	lefty -=1;
        if ( debug) 
	  {
           printf("In find left highy , leftx = %d lefty = %d grey = %d \n",
                         leftx,lefty, pix_out[leftx][lefty].greyval);
	  }
      }

  left_highy = lefty;
  left_highx = leftx;


  if ( debug) 
    {
      printf("In get left_highy , left_highx, left_highy = %d %d \n",
	     left_highx, left_highy );
    }
}

// look thru the out pix data and 
// and find out where the right wide bar ends in the y direction

void find_right_highy(int iny)
{
  int rightx;
  int righty;
  int debug;
  
  debug = 0;

  rightx = save_line_data[save_line_cnt].endlpixel +
    ( save_line_data[save_line_cnt].endwidth / 2);

  righty = iny;

  if (debug)
    {
      printf("In find right highy , rightx, righty = %d %d \n",
	     rightx, righty );
    }

  if (debug) { printf("In find right highy, x,y = %d %d \n", rightx, righty); }

    while(( righty > 0) && ((pix_out[rightx][righty].greyval > dark_threshold)
		   || ( pix_out[rightx][righty-1].greyval > dark_threshold)))
      { 
	righty -=1;
        if (debug)
	  {
           printf("In find right highy , rightx = %d righty = %d grey = %d\n",
                       rightx, righty, pix_out[rightx][righty].greyval);
          }
      }
  right_highy = righty;
  right_highx = rightx;

  if ( debug) 
    {
      printf("In get right_highy , right_highx, right_highy = %d %d \n",
	     right_highx, right_highy );
    }
}

void decode_line()
{
  int ii;
  int end_bar;
  int normal_xy_ind;
  int dnormal_x, dnormal_y, normal_dir, normal_slope;
  int kk;

  int endratio2sp2;
  int endratio3sp3;
  int endratio4sp4;
  int endratio5sp5;

  int end_bar_found;

  int b1wid;

  float b0widthf;
  float b1widthf;
  float b2widthf;
  float b3widthf;
        
  float b1spwidthf;
  float b2spwidthf;
  float b3spwidthf;

  float ratio_1to2;
  float ratio_1to3;
  float ratio_1to4;
  float ratio_2to3;
  float ratio_2to4;
  float ratio_3to4;
  float ratio_1to2sp;
  float ratio_2to2sp;
  float ratio_3to3sp;
  float ratio_4to4sp;
  float fig_of_merit;
  int test_barwidth_low; 
  int test_barwidth_high;
  int fig;
  int debug;

  debug = 0;
  lpixel = 0;
  bar_cnt = 0;

  if ( debug)
    {
      printf("In decode line \n");
    }

  while( lpixel < pixel_cnt )
    {

      skip_underthress();
      if ( lpixel < pixel_cnt)
        {
	  get_bar();
        }
    }

  for ( ii = 0; ii < bar_cnt; ii += 1)
    {
      if (debug)
	{
      printf("Bar data ii = %d width =  %d space_width = %d  lpixel = %d \n",ii,
           bar_array[ii].width, bar_array[ii].space_width,
                    bar_array[ii].lpixel);
        }
    }

  if ( bar_cnt > 2)
    {
      // save away bar information if more than 3 bars
      save_line_data[save_line_cnt].barcnt = bar_cnt;

      if ( bar_cnt < 8)
        {
          end_bar = bar_cnt;
        }
      else
	{
	  end_bar = 8;
        }

        for ( ii = 0; ii < end_bar; ii += 1)
	{
         
	  save_line_data[save_line_cnt].tbar[ii].width = bar_array[ii].width;
          save_line_data[save_line_cnt].tbar[ii].space_width = 
	    bar_array[ii].space_width;
          save_line_data[save_line_cnt].tbar[ii].lpixel = bar_array[ii].lpixel; 
	}

	// calculate merit value

	if ( bar_array[ii].lpixel < b_box_xlen / 2)
	  {
           b0widthf = (float) bar_array[0].width;
           b1widthf = (float) bar_array[1].width;
           b2widthf = (float) bar_array[2].width;
           b3widthf = (float) bar_array[3].width;
        
           b1spwidthf = (float) bar_array[1].space_width;
           b2spwidthf = (float) bar_array[2].space_width;
           b3spwidthf = (float) bar_array[2].space_width;

	   ratio_1to2 = b0widthf / b1widthf;
	   ratio_1to3 = b0widthf / b2widthf;
	   ratio_1to4 = b0widthf / b3widthf;

	   ratio_2to3 = b1widthf / b2widthf;
     	   ratio_2to4 = b1widthf / b3widthf;
	   ratio_3to4 = b2widthf / b3widthf;

	   ratio_2to2sp = b1widthf/b1spwidthf;
	   ratio_3to3sp = b2widthf/b2spwidthf;
	   ratio_4to4sp = b3widthf/b3spwidthf;
           ratio_1to2sp = b1widthf/b2spwidthf;

           onewidthf = ( b1widthf + b2widthf + b3widthf)/3;
          
          
	   fig_of_merit = (float) (fabs( ratio_1to2 - 8.0) +
	      	          fabs( ratio_1to3 - 8.0) +
                         fabs( ratio_1to4 - 8.0) +
                         fabs( ratio_2to3 - 1.0) +
                         fabs( ratio_2to4 - 1.0) +
                         fabs( ratio_3to4 - 1.0) +
                         fabs( ratio_2to2sp - 1.0) +
                         fabs( ratio_3to3sp - 1.0) +
	                 fabs( ratio_4to4sp - 1.0));

	  fig = (int) 100 * fig_of_merit;
          fig = abs( fig );
          }
        else
          {
	    fig = 1000;    // way off value
          }

        if (debug) { printf(" fig = %d \n",fig); }

	save_line_data[save_line_cnt].fig = fig;

        normal_xy_ind = (bar_array[1].lpixel + bar_array[1].width)
	  + (bar_array[1].width/ 4);

        dnormal_x = save_line_data[save_line_cnt].line_x[normal_xy_ind];
        dnormal_y = save_line_data[save_line_cnt].line_y[normal_xy_ind];
        normal_slope = save_line_data[save_line_cnt].slope;

	// dir of normal is negative of line dir
        normal_dir = 0 - save_line_data[save_line_cnt].dir;
  
        // normal slope is the same as line normal, but will
        // be treated in draw normal as x/y not y/x

        // draw normal line , grey data in norm_line_data

	draw_normal(dnormal_x,dnormal_y,normal_slope, normal_dir, b_box_ylen/2);


        normal_xy_ind = (bar_array[0].lpixel + bar_array[0].width/2);

        dnormal_x = save_line_data[save_line_cnt].line_x[normal_xy_ind];
        dnormal_y = save_line_data[save_line_cnt].line_y[normal_xy_ind];
        normal_slope = save_line_data[save_line_cnt].slope;

        if (( norm_fdark_ind == -1) || ( norm_fdark_ind > b_box_ylen/3))
	  {
	    if ( abs( fig ) < 1000)
	      {
                find_height( dnormal_x, dnormal_y, normal_slope,
                                         normal_dir, b_box_ylen);
                normal_xy_ind = ( bar_array[0].lpixel);
                dnormal_x = save_line_data[save_line_cnt].line_x[normal_xy_ind];
                dnormal_y = save_line_data[save_line_cnt].line_y[normal_xy_ind];

                if (debug) { printf("Looking for end bars \n"); }
                end_bar_found = FALSE;

		// try to find the end bars
                for (kk = 3; kk < bar_cnt; kk += 1)
                  {
                    test_barwidth_low = bar_array[0].width - bar_array[1].width
		      - bar_array[1].width - bar_array[1].width;
                    test_barwidth_high = bar_array[0].width +
		      bar_array[1].width * 2;
                    if (debug) {
		      printf("t_barwidth_low = %d , t_barwidth_high = %d \n",
			     test_barwidth_low, test_barwidth_high);
                      printf("bar 0 width = %d bar 1 width = %d \n", 
			     bar_array[0].width, bar_array[1].width );
                      printf("bar kk width = %d for kk = %d \n",
			     bar_array[kk].width, kk );
                  	       }
                           
		    if(( bar_array[kk].width > test_barwidth_low) &&
		       ( bar_array[kk].width < test_barwidth_high))
		      {
                        if (debug) { printf("Test 1 success, kk = %d \n",kk); }
                        if (debug) { printf("bar_cnt = %d \n", bar_cnt); }

			if ( kk < bar_cnt - 4 )  // 4 more bars
                          {
                            endratio2sp2 = 10 * bar_array[kk+1].space_width
			      / bar_array[kk+1].width;
                            endratio3sp3 = 10 * bar_array[kk+2].space_width
			      / bar_array[kk+2].width;
                            endratio4sp4 = 10 * bar_array[kk+3].space_width
			      / bar_array[kk+3].width;
                            endratio5sp5 = 10 * bar_array[kk+4].space_width
			      / bar_array[kk+4].width;
                            if (debug)
			      {
                               printf("endratio2sp2 = %d \n",endratio2sp2);
                               printf("endratio3sp3 = %d \n",endratio3sp3);
                               printf("endratio4sp4 = %d \n",endratio4sp4);
                               printf("endratio5sp5 = %d \n",endratio5sp5);
                              }
                            if (( endratio2sp2 > 3) && ( endratio2sp2 < 18))
                             {
                              if (( endratio3sp3 >18) && ( endratio3sp3 < 49))
                               {
                                if ((endratio4sp4 > 3) && (endratio4sp4 < 18))
                                 {
                                  if ((endratio5sp5 >10)&&(endratio5sp5 < 30))
                                   {
                                    b1wid = bar_array[1].width;
                                    if (( bar_array[kk+1].width/2 ) < b1wid )
                                     {
		       		      if (( bar_array[kk+2].width/2)< b1wid)
					{
					  if ((bar_array[kk+3].width/2) < b1wid)
					  {
                                            if (debug) {
					    printf("End bar found at %d \n",
						 bar_array[kk].lpixel - 
                                                           bar_array[0].lpixel);
					    }
					    end_bar_found = TRUE;
					    
                                     save_line_data[save_line_cnt].endlpixel
                                                   = bar_array[kk].lpixel
					       - bar_array[0].lpixel;
                                     save_line_data[save_line_cnt].has_7111
				                     = TRUE;
                                     save_line_data[save_line_cnt].endwidth=
				      bar_array[kk].width;
                                     save_line_data[save_line_cnt].beginwidth=
				      bar_array[0].width;
					    
                                          }
 					}
                                      }
                                    }
			        }
                              }
                          }
			}
                      }
		   }
		if (( fig < bestfig ) && ( end_bar_found == TRUE))
		  {

                   find_left_edge( dnormal_x, dnormal_y, normal_slope,
                        normal_dir, b_box_ylen/2 , b_box_xlen);

		   if (debug) { printf("Found left_edge finished \n"); }

                    fig_flag = TRUE;
		    best_normal_x = dnormal_x;
                    best_normal_y = dnormal_y;
                    best_normal_slope = normal_slope;
                    best_normal_dir = normal_dir;
   		    bestfig = fig;
                   onewidth = ( int) onewidthf;  // width of 1 unit
		   best_onewidth = onewidth;
                   if ( debug) {
		     printf("Setting best_onewidth = %d \n", best_onewidth); }

                   begin_left_col = bar_array[4].lpixel- bar_array[0].lpixel;
                   begin_right_bars = save_line_data[save_line_cnt].endlpixel;
                  
                   best_begin_left_col = begin_left_col;
                   best_begin_right_bars = begin_right_bars;
                   num_sym_cols = (  begin_right_bars - begin_left_col)/
		     begin_left_col;
                   best_num_sym_cols = num_sym_cols;

                   onewf = (float) ( begin_left_col);
                   onewf = onewf / 17;

                   best_onewf = onewf;
                   if (debug)
		     {
                      printf("Setting best_onewf = %f \n", best_onewf);
                     }

                   if (debug) { printf("About to call find_left_highy \n"); }

                   find_left_highy(b_box_ylen/2);
                   find_right_highy(b_box_ylen/2);

                              
                   if ( (right_highx - left_highx) != 0)
		     {
                      pix_slope = ( 1000 * ( right_highy - left_highy)) /
		        ( right_highx - left_highx);
                     }
                   else
                     {
                       printf("No change in x direction for symbol \n");
                       exit(0);
                     }
                   best_pix_slope = pix_slope;
                   best_pixel_out_xlen = pixel_out_xlen;
                   best_pixel_out_ylen = pixel_out_ylen;

                   final_pix( pixel_out_xlen, pixel_out_ylen);


                   if (debug) {
                    printf("New best fig = %d  onewidth = %d \n",fig,onewidth);
		   }

                  }               
              }
          }

      save_line_cnt += 1;

    } // more than 3 bars

}



//   threshold the line data
//   
//
void smooth_line()
{
  int jj;

  int pixel_new;
  int debug;


  debug = 0;

  for(jj = 0; jj < pixel_cnt; jj += 1)
    {
      if ( line_data[jj] > line_threshold )
        {
          line_data[jj] = 127;
        }
      else
	{
	  line_data[jj] = 0;
        }
    }

  
  if ( debug) 
    {
      printf("End smooth line \n");
    }
}

/**************************************************************************
 *  line_fast                                                             *
 *    gets a line using Bresenham's line-drawing algorithm, which uses   *
 *    no multiplication or division.                                      *
 *    gets data from pixel array and puts in line_data                    *
 **************************************************************************/

void line_fast(int x1, int y1, int x2, int y2)
{
  int i,dx,dy,sdx,sdy,dxabs,dyabs,x,y,px,py;

  dx=x2-x1;      /* the horizontal distance of the line */
  dy=y2-y1;      /* the vertical distance of the line */
  dxabs=abs(dx);
  dyabs=abs(dy);
  sdx=sgn(dx);
  sdy=sgn(dy);
  x=dyabs>>1;
  y=dxabs>>1;
  px=x1;
  py=y1;

  pixel_cnt = 0;

  get_pixel(px,py);

  if (dxabs>=dyabs) /* the line is more horizontal than vertical */
  {
    for(i=0;i<dxabs;i++)
    {
      y+=dyabs;
      if (y>=dxabs)
      {
        y-=dxabs;
        py+=sdy;
      }
      px+=sdx;
      get_pixel(px,py);
    }
  }
  else /* the line is more vertical than horizontal */
  {
    for(i=0;i<dyabs;i++)
    {
      x+=dxabs;
      if (x>=dyabs)
      {
        x-=dyabs;
        px+=sdx;
      }
      py+=sdy;
      get_pixel(px,py);
    }
  }
}

/**************************************************************************
 *  line_fast_post                                                       *
 *    gets a line using Bresenham's line-drawing algorithm, which uses   *
 *    no multiplication or division.                                      *
 *   get data from the pix_out and puts it in post_line_data              *
 **************************************************************************/

void line_fast_post(int x1, int y1, int x2, int y2)
{
  int i,dx,dy,sdx,sdy,dxabs,dyabs,x,y,px,py;

  dx=x2-x1;      /* the horizontal distance of the line */
  dy=y2-y1;      /* the vertical distance of the line */
  dxabs=abs(dx);
  dyabs=abs(dy);
  sdx=sgn(dx);
  sdy=sgn(dy);
  x=dyabs>>1;
  y=dxabs>>1;
  px=x1;
  py=y1;

  post_pixel_cnt = 0;

  get_pixel_post(px,py);

  if (dxabs>=dyabs) /* the line is more horizontal than vertical */
  {
    for(i=0;i<dxabs;i++)
    {
      y+=dyabs;
      if (y>=dxabs)
      {
        y-=dxabs;
        py+=sdy;
      }
      px+=sdx;
      get_pixel_post(px,py);
    }
  }
  else /* the line is more vertical than horizontal */
  {
    for(i=0;i<dyabs;i++)
    {
      x+=dxabs;
      if (x>=dyabs)
      {
        x-=dyabs;
        px+=sdx;
      }
      py+=sdy;
      get_pixel_post(px,py);
    }
  }
}

/**************************************************************************
 *  line_fast norm                                                      *
 *    gets a line using Bresenham's line-drawing algorithm, which uses   *
 *    no multiplication or division.                                      *
 *    gets data from pixel_array but puts result in norm_line_data        *
 **************************************************************************/

void line_fast_norm(int x1, int y1, int x2, int y2)
{
  int i,dx,dy,sdx,sdy,dxabs,dyabs,x,y,px,py;

  dx=x2-x1;      /* the horizontal distance of the line */
  dy=y2-y1;      /* the vertical distance of the line */
  dxabs=abs(dx);
  dyabs=abs(dy);
  sdx=sgn(dx);
  sdy=sgn(dy);
  x=dyabs>>1;
  y=dxabs>>1;
  px=x1;
  py=y1;

  pixel_cnt = 0;

  get_pixel_norm(px,py);

  if (dxabs>=dyabs) /* the line is more horizontal than vertical */
  {
    for(i=0;i<dxabs;i++)
    {
      y+=dyabs;
      if (y>=dxabs)
      {
        y-=dxabs;
        py+=sdy;
      }
      px+=sdx;
      get_pixel_norm(px,py);
    }
  }
  else /* the line is more vertical than horizontal */
  {
    for(i=0;i<dyabs;i++)
    {
      x+=dxabs;
      if (x>=dyabs)
      {
        x-=dyabs;
        px+=sdx;
      }
      py+=sdy;
      get_pixel_norm(px,py);
    }
  }
}

// draw a normal at a point

void draw_normal(int nx, int ny, int nslope, int ndir, int ylen)
{
  int ii;
  int xinc, yinc;
  int dark_percent;

  int bits_on_cnt;
  int num_dark;
  int first_dark;
  int first_dark_ind;
  int debug;

  debug = 0;

  if (debug)
    {
     printf("In draw normal , x,y = %d %d \n",nx, ny );
     printf("Slope = %d \n", nslope);
     printf("ylen = %d \n", ylen);
     printf("dir = %d \n",ndir);
    }


  if ( ndir == 1)  // positive x and positive y incr
    {
       yinc = (ylen);
       xinc = (((ylen) * slope) / 100);
      
      line_fast_norm(nx, ny, nx + xinc, ny+yinc);
    }

  if ( ndir == -1)   // negative x incr and positive y incr
    {
       yinc = (ylen);
       xinc = 0- (( (ylen) * slope) / 100);

      line_fast_norm(nx, ny, nx + xinc, ny+yinc);
    }

  num_dark = 0;

  first_dark = FALSE;
  first_dark_ind = -1;

  for ( ii = 0; ii < pixel_cnt; ii += 1)
    {
      if ( norm_line_data[ii] > dark_threshold)
	{
          if ( first_dark == FALSE)
	    {
	      first_dark = TRUE;
              first_dark_ind = ii;
            }
         
	  num_dark += 1;
        }
      if (debug) {  printf("norm line %d = %d \n",ii, norm_line_data[ii]); }

    }

  if (debug)
    {
      printf("num_dark = %d \n",num_dark);
      printf("First dark at = %d \n",first_dark_ind);
    }

  dark_percent = (num_dark * 100) / pixel_cnt;

  if (debug) { printf("Dark percent = %d \n", dark_percent); }

  norm_fdark_ind = first_dark_ind;
  
}


 // use a binary search to find the index in the
 // bit_array of the pdf barcode
 //
 int find_index( long inbits, int code_in, int column, int row)
 {

   int ii;
   int found_index;
   int bit_index;
   int bit_increment;
   int found;
   int less_than_flag;

   int debug;

   bit_index = 1392;

   bit_increment = 696;

   found = FALSE;

   less_than_flag = FALSE;

   debug = 0;

   if ( debug) { printf("Inbits = %x \n", inbits); }

   while (( found == FALSE ) && (bit_increment > 11 ))
     {
       if (debug)
	 {
	     printf(" bit_array val = %x , inbits = %x \n",
				  bit_array[bit_index], inbits); 
	 }
      if (bit_array[bit_index] > inbits)
	{
	 bit_index = bit_index - bit_increment;
	 less_than_flag = FALSE;
	}
      else
	{
	  if (bit_array[bit_index] == inbits)
	    {
	      found = TRUE;
	      found_index = bit_index;
	    }
	  else  // less than 
	    {
	     bit_index = bit_index + bit_increment;
	     less_than_flag = TRUE;
	    }
	}

      bit_increment = bit_increment / 2;

      if (debug)
	{
	 printf("In find loop , bit_increment = %d \n", bit_increment);
	 printf("In find loop , bit_index = %d \n", bit_index);
	}

     }

  if ( found == FALSE)
    {
      ii = 0;
      if ( debug) { printf("no less than flag , bit_index = %d \n",bit_index); }
      while (( found == FALSE) && ( ii < 25) && ( bit_index < 2759))
	{
	   if (bit_array[bit_index + ii] == inbits)
	     {
	      found = TRUE;
	      found_index = bit_index + ii;
	     }
	   ii += 1;        
	}
    }

  if ( found == FALSE)
    {
      ii = 0;
      if (debug) { printf("less than flag , bit_index = %d \n",bit_index); }
      while (( found == FALSE) && ( ii > -25) && ( bit_index > 26))
	{
	  if ( bit_index + ii > 0)
	   {
	    if (bit_array[bit_index + ii] == inbits)
	     {
	      found = TRUE;
	      found_index = bit_index + ii;
	     }
	   }
	   ii -= 1;        
	}
    }

  if (found == FALSE)
    {

      // do an exhaustive search

     ii = 0;

     if (debug) { printf("Doing exhaustive search \n"); }

     while ( ( ii < 2784 ) && (found == FALSE))
       {
	  if ( bit_array[ii] == inbits)
	    {
	     found = TRUE;
	     found_index = ii;
	    }
	   ii += 1;
       }
    }

    if (found == FALSE)
      {
	if (code_in == 0)
	  {
           printf(" Top left module not found in bit_array = 0x%x \n", inbits);
          }

	if (code_in == 1)
	  {
           printf(" Bottom left module not found in bit_array = 0x%x \n", inbits);
          }
	if (code_in == 2)
	  {
	    if ( debug) 
              {
               printf(" Column,row module not found in bit_array = 0x%x \n", inbits);
               printf(" Column = %d row = %d\n", column, row);
              }
          }
       return(-1);
      }
    else
      {
       return( found_index );
      }

 }


// find the differences in the x ( row ) grey values 
//  for each y location in pix_col data
//

void do_pixcol_diffy(int inwidth, int inheight)
{
  int ii;
  int jj;
  int kk;
  int diff_data[OUT_MAX_X_PIXEL];
  int diff_tots[OUT_MAX_X_PIXEL];
  int first_deriv[OUT_MAX_X_PIXEL];
  int diff_tots_sorted[OUT_MAX_X_PIXEL];
  int total_diff;
  int prev_total;
  long cum_total;
  long avg_total;
  int prev_line;
  int x_space;
  int est_bar_width;
  int debug;


  debug = 0;

  diff_data[0] = 0;
  diff_data[1] = 0;
  diff_data[2] = 0;

  cum_total = 0;
  total_diff = 0;
  prev_total = 0;
  if (debug)
    {
       printf(" In do pix_col_diffy, inwidth, inheight = %d %d \n",
                           inwidth, inheight);
    }
  for ( jj = 3; jj < inwidth; jj += 1)
    {
      total_diff = 0;
      for ( ii = 0; ii < inheight ; ii += 1)
	{
      diff_data[ii] = abs(pix_cols[jj][ii].greyval - pix_cols[jj-3][ii].greyval);
      total_diff = total_diff + diff_data[ii];
        }
      diff_tots[jj]= total_diff;
      cum_total = total_diff + cum_total;
      first_deriv[jj] = total_diff - prev_total;
      prev_total = total_diff;
     }  

  if ( inwidth != 0)
    {
     avg_total = cum_total/inwidth;
    }
  else
    {
      printf("In pixcol_diff_y, inwidth = 0 , internal error \n");
      exit(0);
    }


  if (debug)
    { printf("avg total = %d \n", avg_total); }

  for ( jj = 0; jj < inwidth; jj += 1)
    {
      if (debug)
        {
      printf(" jj = %d diff_tots = %d 1st deriv = %d \n",jj, diff_tots[jj],
                first_deriv[jj]);
        }
    }

  kk = 0;
  prev_line = 0;
  for ( jj = 5; jj < inwidth; jj += 1)
    {
      if (( first_deriv[jj] > 0) && ( first_deriv[jj+1] < 0))
        {
          if ( (long) diff_tots[jj] > (avg_total/2 ))
            {
             
              if ( kk < MAX_LINES_X)
		{
		  if ( kk > 0)
                    {
                     if (( jj - pix_col_xline[kk-1]) > (int) (best_onewf /2))
		      {
		       pix_col_xline[kk] = jj;
                       x_space = jj - prev_line;
                       prev_line = jj;  
                       kk += 1;
		      }
                    }
                  else
                    {
                    if ( jj  > (int) (best_onewf /2))
		      {
		       pix_col_xline[kk] = jj;
                       x_space = jj - prev_line;
                       prev_line = jj;  
                       kk += 1;
		      }
                    }
                }
              else
		{
		  printf("Too many vertical lines in pix_col \n");
                }
              if (debug)
                {
                 printf("Candidate at jj = %d  x_space = %d \n",jj, x_space);
                }
            }
        }
    }
    
  vert_line_cnt = kk;

  if ( kk != 0)
    {
     est_bar_width = (inwidth * 10 )/ kk;
    }
  else
    {
      est_bar_width = 1;
    }

  if (debug)
    {
     printf("est_bar_width = %d \n", est_bar_width);
     for ( kk = 0; kk < vert_line_cnt; kk += 1) 
       {
         printf("kk = %d pix_col_xline = %d \n", kk , pix_col_xline[kk]);
       }
    }

}



// find the differences in the x ( row ) grey values 
//  for each y location in pix_col data
//

void do_pixcol_diffx(int inwidth, int inheight)
{
  int ii;
  int jj;
  int kk;
  int diff_data[OUT_MAX_X_PIXEL];
  float diff_tots[OUT_MAX_Y_PIXEL];
  float first_deriv[OUT_MAX_Y_PIXEL];
  float total_diff;
  float prev_total;
  float cum_total;
  long avg_total;
  int prev_line;
  int y_space;
  int pix_ydiffs_tot;
  int debug;
  int avg_yspace;

  debug = 0;
  y_space = 0;

  diff_data[0] = 0;
  diff_data[1] = 0;
  diff_data[2] = 0;

  cum_total = 0.0;
  total_diff = 0.0;
  prev_total = 0.0;

  if (debug) { printf("In do_pixcols_diffx , inwidth = %d , inheight = %d \n",
		      inwidth, inheight); }

  for ( jj = 3; jj < inheight; jj += 1)
    {
      total_diff = 0.0;
      for ( ii = 0; ii < inwidth ; ii += 1)
	{
	  if ( ii < OUT_MAX_X_PIXEL)
	    {
            diff_data[ii] = abs(pix_cols[ii][jj].greyval
                                    - pix_cols[ii][jj-3].greyval);
            }
          else
	    {
	      printf("diff_data array size exceed \n");
              exit(0);
            }
          
          total_diff = total_diff + (float) diff_data[ii];
        }
      if ( jj < OUT_MAX_Y_PIXEL)
	{
         diff_tots[jj]= total_diff;
         cum_total = total_diff + cum_total;
         first_deriv[jj] = total_diff - prev_total;
         prev_total = total_diff;
        }
      else
        {
	  printf("diff_tots array size exceeded \n");
          exit(0);
        }
     }  

  if (debug) { printf("Completed j = 3 to inheight loop  -inheigth = %d\n",
                     inheight); }

  if ( inheight != 0)
    {
     avg_total = (int) (cum_total/(float) inheight);
    }
  else
    {
      printf("In diff_col_pix_x , inheight = 0, internal error \n");
      exit(0);
    }

  if (debug) { printf("avg_total = %d \n", avg_total); }
  if (debug)
    {
     printf("avg total = %d \n", avg_total);
    }
  for ( jj = 0; jj < inheight; jj += 1)
    {
      if (debug)
	{
         printf(" jj = %d diff_tots = %f 1st deriv = %f \n",jj, diff_tots[jj],
                first_deriv[jj]);
        }
    }

  kk = 0;
  prev_line = 0;
  pix_ydiffs_tot = 0;
  for ( jj = 5; jj < inheight; jj += 1)
    {
      if (( first_deriv[jj] > -1.0) && ( first_deriv[jj+1] < 0.0))
        {
          if (  diff_tots[jj] > (avg_total  * 2))
            {
             
              if ( kk < 300)
		{
		  pix_col_yline[kk] = jj;
                  if ( debug)
		    { printf("Added to pix_col_yline ,kk, val = %d %d \n",
			     kk, pix_col_yline[kk]); }
                  if ( kk > 1)
                    {
		      pix_col_ydiffs[kk-1] = pix_col_yline[kk] -
                                              pix_col_yline[kk-1];
                      pix_ydiffs_tot += pix_col_ydiffs[kk-1];
                    }
                      
                  y_space = jj - prev_line;
                  prev_line = jj;  
                  kk += 1;
                }
              else
		{
		  printf("Too many horizontal lines in pix_col \n");
                }
              if (debug)
		{
                 printf("Candidate at jj = %d  y_space = %d \n",jj, y_space);
		}
            }
        }
    }

  if ( kk > 1)
    {
     avg_yspace = (pix_ydiffs_tot / (kk-1));
     if ( debug) {  printf("avg_yspace = %d \n", avg_yspace); }
    }
  if (debug) { printf("Completed j = 5 to inheight loop \n"); }
    
  if ( kk != 0)
    {
     est_bar_height = (inheight  )/ kk;
    }
  else
    {
      est_bar_height = 4;
      printf("Possible problem with bar height estimate, using 4 \n");
    }
 
  if (debug)
    {
     printf("est_bar_height = %d \n", est_bar_height);
    }

}

int test_black( int inx, int iny, int near)
{

  int tot;
  int avg_grey;

  if ( near == 1)
    {
      tot = 0;
      tot = tot + pix_cols[inx][iny].greyval;
      tot = tot + pix_cols[inx+1][iny].greyval;
      tot = tot + pix_cols[inx-1][iny].greyval;

      tot = tot + pix_cols[inx][iny-1].greyval;
      tot = tot + pix_cols[inx+1][iny-1].greyval;
      tot = tot + pix_cols[inx-1][iny-1].greyval;

      tot = tot + pix_cols[inx][iny+1].greyval;
      tot = tot + pix_cols[inx+1][iny+1].greyval;
      tot = tot + pix_cols[inx-1][iny+1].greyval;

      avg_grey = tot/9;
    }

  if ( near == 0)
    {
   
      tot =  pix_cols[inx][iny].greyval;
   
      avg_grey = tot;
    }

  if ( avg_grey < dark_threshold)
    {
      return(0);
    }
  else
    {
      return(1);
    }

}

int  bits_to_symbol( int *inbits)
{

  int ii;
  int mask;
  int outsym;

  ii = 0;
  mask = 1 << 16;
  outsym = 0;

  while (ii < 17)
    {
      if ( inbits[ii] == 1)
        {
	  outsym = outsym | mask;         
        }
      ii += 1;
      mask = mask >> 1;

    }
  return(outsym);
}

// get the top most left bars ( a 17 bar symbol)
// this will give hints about actual number of rows

void do_top_bar(int bar_height_in)
{

  int ii;
  int testx;
  int testy;
  int onewid;
  int bit[20];
  int bit_int;  // integer rep of bits
  int bitcode;
  int rowval;
  int rowind;
  int numrowsm1_div3;
  int numrow_est;
  int numrow_est2;
  int top_bar_index;
  int bot_bar_index;
  int ecc_val;
  int this_row;
  int num_col_est;
  int row_height_estlo;
  int row_height_esthi;
  int numrows;          // value gotten from both top and bottom
  int this_row_height;
  int init_testx;
  float base;
  int debug;

  debug = 0;

  if (debug)
    {
     printf("In do top bar \n");
    }

  onewid = (int) best_onewf; 
  //  testy = pix_col_yline[0]/ 2;
  
  testy = (est_bar_height / 2);
  
 
  y_begin = testy;
  base = ( best_onewf / 2);

  // get the top module from column one

  for (ii=0; ii < 17 ; ii += 1)
    {
      testx = (int) (( (float)ii * best_onewf)  + base);
      if (debug) {
	printf("In do top bar, testx , testy = %d %d \n", testx, testy); }
      bit[ii] = test_black( testx, testy, 1);
    }

  for (ii = 0; ii < 17; ii += 1)
    {
      if (debug) { printf("ii = %d bit[ii] = %d \n", ii, bit[ii]); }
    }

     bit_int = bits_to_symbol( bit );

     if (debug) { printf("bit_int = %x \n", bit_int); }

     top_bar_index = find_index(bit_int,0,0,0);

     if ( top_bar_index != -1 )
       {
         if (debug) { printf("Top bar index = %d \n", top_bar_index); }

	 bitcode = bit_codes[top_bar_index]-1;
         rowval = bitcode / 929;
         if ( rowval != 0)
           {
	     printf("Error: Top left bar code is not a row 0 code \n");
             exit(-1);
           }
         rowind = (bitcode % 929) / 30;
         numrowsm1_div3 = ( bitcode % 929) % 30;
         numrow_est = (3 * numrowsm1_div3) + 1;
         if (debug) { printf("Numrow_est = %d-%d\n", numrow_est,numrow_est+2);}
         row_height_estlo = ( bar_height_in / ( numrow_est + 2));
         row_height_esthi =  bar_height_in / numrow_est;
         if (debug)
	   {
             printf("bar_height_in = %d row_height_estlo, esthi = %d %d \n",
                 bar_height_in, row_height_estlo,
		row_height_esthi);        
           }
         row_height_avg = ( row_height_estlo + row_height_esthi) / 2;
       }
     else
       {
         printf("Top left bar code is not a correct bar \n");
       }

     if (debug) { printf("top_bar_index = %d \n", top_bar_index); }

    
     // testy = bar_height_in - (pix_col_yline[0]/ 2);

     // testy = total height of all bars - height of one bar /2

     testy = bar_height_in - (est_bar_height/2);
     init_testx = onewid /2 ;

     bot_bar_index = -1;
    
     base = ( best_onewf / 2);
     while(( bot_bar_index == -1 ) && ( base > -1 ))
       {
        
        // get the bottom bar code symbol

         for (ii=0; ii < 17 ; ii += 1)
          {
           testx = (int) (( (float)ii * best_onewf)  + base);
           // testx = testx + onewid;
           if (debug) {
             printf("In do left bot  bar, testx , testy = %d %d \n", 
                             testx, testy);
	   }
           
           bit[ii] = test_black( testx, testy, 0);
           if ( debug ) { printf("bit = %d \n",bit[ii]); }
          }

         bit_int = bits_to_symbol( bit );
         if (debug) { printf("bit_int for bottom symbol  = %x \n", bit_int); }
         base -=1;
        bot_bar_index = find_index(bit_int,1,0,0);
       }

     if ( bot_bar_index != -1 )
       {
         if (debug) { printf("Bottom bar index = %d \n", bot_bar_index); }

	 bitcode = bit_codes[bot_bar_index]-1;
         rowval = bitcode / 929;
         
         if (debug) {
            printf(" bottom bitcode = %d , rowval = %d \n", bitcode, rowval);
	 }
         if ( rowval == 0)
	   {
            rowind = (bitcode % 929) / 30;
            this_row = ( 3 * rowind) -1;
            numrowsm1_div3 = ( bitcode % 929) % 30;
            numrow_est2 = (3 * numrowsm1_div3) + 1;
            numrows = numrow_est;
            best_row_height = (bar_height_in / numrows);
            if (debug) { printf("best row height = %d \n", best_row_height); }
           }

         if ( rowval == 1)
	   {
            rowind = (bitcode % 929) / 30;
            this_row = ( 3 * rowind) -1;
            ecc_val = ( bitcode % 929) % 30;
           
            numrows = numrow_est + 1;
            best_row_height = ( bar_height_in / numrows);
            if (debug) { printf("best row height = %d \n", best_row_height); }
           }

         if ( rowval == 2)
	   {
            rowind = (bitcode % 929) / 30;
            this_row = ( 3 * rowind) -1;
            num_col_est = (( bitcode % 929) % 30) + 1;
            if (debug) {
              printf("Number of columns from left bars = %d \n", num_col_est);}

            numrows = numrow_est + 2;
            best_row_height = ( bar_height_in/ numrows);
            if (debug) { printf("best row height = %d \n", best_row_height); }
           }
         if ( numrows != this_row)
	   {
             if ( debug)
	       {
	        printf("Mismatch between row number calculations \n");
                printf("numrows = %d this_row = %d \n", numrows, this_row);
               }
           }
         else
	   {
             if (debug) { printf("numrows = %d \n", numrows); }
             this_row_height =  bar_height_in /  numrows;
      
	     if (debug) { printf("good row height = %d \n", this_row_height); }
            
           }

       }
     else
       {
         printf("Bottom left bar code is not a correct bar \n");
       }

}

// get symbols in an entire column
// 

void do_col_bars(int inrows)
{

  int ii;
  int jj;
  int testx;
  int testy;
  int onewid;
  int bit[20];
  int bit_int;
  int symbs[300];
  int debug;

  debug = 0;

  for ( jj = 0; jj < inrows; jj += 1)
    {
     onewid = (int) onewf; 
     testy = pix_col_yline[jj] - (est_bar_height/20);
     testx = 0;
     for (ii=0; ii < 17 ; ii += 1)
       {
        testx = testx + onewid;
        if (debug) {
          printf("In do col bars, testx , testy = %d %d \n", testx, testy); }

         bit[ii] = test_black( testx, testy, 1);
        }

     for (ii = 0; ii < 17; ii += 1)
       {
	 if (debug) { printf("ii = %d bit[ii] = %d \n", ii, bit[ii]); }
       }

     bit_int = bits_to_symbol( bit );


     if ( jj < 300)
       {
	 symbs[jj] = bit_int;
       }
    }
}

// look thru the unslanted pix final data 
// and get all the x values where columns may start
//

void get_columns_start( int ylenin)
{
  int ii, jj, kk;
  int beginx;
  int endx;
  float col_avg;
  int bit_diff;
  float col_diff;
  float col_tot;
  float tot_l_avg;
  float tot_r_avg;
  int tot_lr_avg;
  int min_avg;
  int prev_lr_avg; 
  int tot_lr_incr;
  int col_cnt;
  int col_start;
  int col_incr;
  int debug;
  float lr_float_avg;
  float tot_lr_float;
  int lr_avg;

  int colcnt;
  

  debug = 0;

  
  if (debug) { printf("In  get columns start, begin_left_col - %d\n",
                      begin_left_col); }

  //beginx = 16 * onewidth;
  beginx = best_begin_left_col;
  endx = best_begin_right_bars;

  min_avg = 10000;
  prev_lr_avg = 0;

  tot_lr_float = 0.0;

  for ( ii = beginx - onewidth; ii < endx; ii += 1)
    {
      // look at all the columns behind this x up to onewidth
      // at the beginning of the column, this should be all white
      
      col_tot = 0.0;
      for ( kk = 0 - onewidth; kk < 0 ; kk += 1)
	{
          col_diff = 0.0;
          for ( jj = 0; jj < ylenin; jj += 1)
	    {
             bit_diff = abs( pix_final[ii+kk][jj].greyval - 0);
             col_diff = col_diff + (float) bit_diff;
             if ( ii == beginx )
               {
		 //        if (debug)
		 // { printf("pix_final=%d \n", pix_final[ii+kk][jj].greyval); }
               }
            }
          col_avg =  (col_diff / (float) jj );
          col_tot = col_avg + col_tot;
         }

      tot_l_avg = col_tot / (float) onewidth;

      // look at all the columns ahead of this x up to onewith
      // at the beginning of the column this should be all black
      
      col_tot = 0.0;
      for ( kk = 0; kk < onewidth ; kk += 1)
	{
          col_diff = 0.0;
          for ( jj = 0; jj < ylenin; jj += 1)
	    {
             bit_diff = abs( pix_final[ii+kk][jj].greyval - 255);
             col_diff = col_diff + (float) bit_diff;
             if ( ii == beginx )
               {
		 // if (debug) { 
		 // prinptf("pix_final b = %d jj = %d kk = %d\n", 
		 //     pix_final[ii+kk][jj].greyval, jj, kk); }
               }
            }
          col_avg = col_diff / (float) jj;
          col_tot = col_avg + col_tot;
        }

      tot_r_avg = col_tot / (float) onewidth;

      tot_lr_avg = (int) (tot_l_avg + tot_r_avg);

      if ( ii < MAX_TOT_LR)
	{
         tot_lr[ii] = tot_lr_avg;
         tot_lr_float = tot_lr_float + ( float) tot_lr_avg;
         tot_lr_incr = tot_lr_avg - prev_lr_avg;
         tot_lr_in[ii] = tot_lr_incr;
        }
      else
	{
          printf("Size of tot_lr = %d pixels exceeded \n", ii);
          exit(-1);
        }

      if ( debug)
	{
       printf(" in columns start, ii = %d  lr = %d , l = %f r = %f incr=%d\n",
             ii , tot_lr_avg, tot_l_avg, tot_r_avg, tot_lr_incr);
        }

      prev_lr_avg = tot_lr_avg;
      if ( tot_lr_avg < min_avg )
        {
          min_avg = tot_lr_avg;
        }
      colcnt += 1;
    }

  lr_float_avg = tot_lr_float / (float) colcnt;

  
  lr_avg = ( int) lr_float_avg;


  if ( debug){ printf("min_avg = %d \n",min_avg); }

  col_cnt = 0;
  for ( ii = beginx - onewidth; ii < endx - (6 * onewidth); ii += 1)
    {
      if (debug) { printf(" ii = %d tot_lr for ii = %d \n",ii, tot_lr[ii]); }
      if (debug) { printf(" tot_lr_in[ii] = %d to_lr_in[ii+1] = %d \n",
			  tot_lr_in[ii], tot_lr_in[ii+1]); }

      if ( (tot_lr[ii] - (min_avg/5)) < min_avg)  // within 20% of min
	{
	  // minimum point according to derivative changing signs
          if (( tot_lr_in[ii] < 1) && ( tot_lr_in[ii+1] > 0))
	    {
	      if ( debug) { printf("near minimum at ii = %d \n",ii); }
             if ( col_cnt < 200)
	       {
                 col_begin_x[col_cnt] = ii;
                 col_cnt += 1;
               }
	     else
               {
                 printf("Too many columns to fit in array of 200\n");
               }
            }
        }
    }  

  if ( col_cnt > 1)
    {
     best_col_width_avg = ( col_begin_x[col_cnt-1] - col_begin_x[0]) / (col_cnt-1);
    }
  else
    {
      best_col_width_avg = 17 * onewidth;
    }

  if (debug) { printf("best_col_width_avg = %d \n", best_col_width_avg); }

    if ( (col_cnt) == num_sym_cols)
    {
      // need checks here for regular spacing etc

      wr_data_file = fopen("out.data","w");

      if (debug)
	{
         printf(" Number of columns agree!!! \n");
        }

      for ( jj = 0; jj < col_cnt; jj +=1 )
        {
	  
          get_a_column( jj, col_begin_x[jj], best_onewidth);
        }

      fclose(wr_data_file);
 
    }
  else
    {
      if ( col_cnt < num_sym_cols)
        {

      printf("Interpolate missing columns - col_cnt = %d num_sym_cols = %d\n",
                   col_cnt, num_sym_cols);

      wr_data_file = fopen("out.data","w");

      if (debug)
	{
         printf(" Number of columns do not agree!!! \n");
         printf("col_cnt = %d num_sym_cols = %d  \n", col_cnt, num_sym_cols);
        }

      col_start = best_begin_left_col;
      col_incr = (best_begin_right_bars - best_begin_left_col ) / num_sym_cols;
      for ( jj = 0; jj < num_sym_cols; jj +=1 )
        {
	  
          get_a_column( jj, col_start, best_onewidth);
          col_start += col_incr;
          
          if (debug) { printf("Column start = %d \n", col_start); }
          
        }

        fclose(wr_data_file);

        }
      else
        {
	  printf("Too many columns detected \n");
          exit(0);
        }
    }

}

// pix_final may be slanted, fix the slant
//
void unslant_pix_final(int xlenin, int ylenin)
{

  int ii;
  int top_bar_width;
  int bot_bar_width;
  int slant;
  int addbits;
  int jj;
  int top_index;
  int bot_index;
  int ht;
  int debug;

  debug = 0;

  if ( debug)
    {
      printf("In unslant pix_final \n");
    }

  top_index = 0;

  while((pix_final[(3 * onewidth)][top_index].greyval < dark_threshold) &&
	( top_index < ylenin))
    {
      top_index += 1;
    }
        
  top_index += 2;

  if (debug) { printf("unslant : top_index = %d \n", top_index); }
  ii = onewidth;
  while ((pix_final[ii][top_index+2].greyval > dark_threshold) && ( ii < xlenin))
    {
      ii += 1;
    }

  top_bar_width = ii;

  bot_index = ylenin;

  while((pix_final[(3 * onewidth)][bot_index].greyval < dark_threshold) &&
	( bot_index > 0))
    {
      bot_index -=1;
    }
  
  bot_index -= 2;
  
  if (debug) { printf("unslant : bot_index = %d \n", bot_index); }

  ii = onewidth;
  while (( pix_final[ii][bot_index].greyval > dark_threshold) && ( ii<xlenin))
    {
      ii += 1;
    }

  bot_bar_width = ii;

  ht = bot_index - top_index;

  if (debug) 
    {
     printf("unslant: top left bar width = %d \n", top_bar_width);
     printf("unslant: bottom left bar width = %d \n", bot_bar_width);
     printf("unslant: slant = %d \n", slant);
    }

  if ( top_bar_width > bot_bar_width)
    {
      slant = top_bar_width - bot_bar_width;

       if (debug) { printf("unslant: slant = %d \n", slant); }

       if ( slant > (onewidth/4))
	 {
          for (ii =top_index; ii < bot_index+2; ii += 1)
	   {
            addbits = ((ii - top_index) * slant)/ ht;
            for (jj = xlenin; jj > addbits -1; jj -=1)
             {
              pix_final[jj][ii].greyval = pix_final[jj-addbits][ii].greyval;
            
             }
            for( jj = addbits -1; jj > -1; jj -=1)
             {
	      pix_final[jj][ii].greyval = 255;
             }
	   }
        }
    }

  else     // slant is opposite
    {
      slant =  bot_bar_width - top_bar_width;;

       if (debug) { printf("unslant: slant = %d \n", slant); }

       if ( slant > (onewidth/4))
	 {
          for (ii =bot_index; ii < top_index-2; ii -= 1)
	   {
            addbits = (( bot_index -ii) * slant)/ ht;
            for (jj = xlenin; jj > addbits -1; jj -=1)
             {
              pix_final[jj][ii].greyval = pix_final[jj-addbits][ii].greyval;
            
             }
            for( jj = addbits -1; jj > -1; jj -=1)
             {
	      pix_final[jj][ii].greyval = 127;
             }
	   }
        }
    }
}

// convert bars to bits 
// 
//  returns total width of bar 
//    
int mod_bar_to_mod_bit(float roundval)
{
  int kk;
  int tot_width;
  int ii,jj;
  int bwwidth;
  int wwidth;
  int bwidth;
  int debug;
  
  debug = 0;

  kk = 0;
  tot_width = 0;
  for ( ii = 0; ii < mod_bar_cnt; ii += 1)
    {
      bwidth = ( (int) (((float) mod_bar_array[ii].width / best_onewf)
                               + roundval));
      if ( bwidth < 7)
        {
          for ( jj = 0; jj < bwidth; jj += 1)
	    {
	      if (( kk > -1) && ( kk < 17))
		{
	         mod_bit_array[kk] = 1;
                  kk = kk + 1;
                }
            }
        }
      else
        {
	  printf(" Module has dark bar that is too wide - width = %d \n",
		 bwidth);
        }

      if ( ( ii + 1) < mod_bar_cnt )
        {
        wwidth=((int)((( float) mod_bar_array[ii+1].space_width / best_onewf)
		      +  roundval ));
                                     
        }
      else
        {
          wwidth = (int) (((float)
			   (mod_pixel_cnt - mod_bar_array[ii].endx) / onewf)
	    + roundval );
        }
       if ( wwidth < 7)
	 {      
          for ( jj = 0; jj < wwidth; jj += 1)
	    {
	      if (( kk > -1) && ( kk < 17))
		{
	         mod_bit_array[kk] = 0;
                  kk = kk + 1;
                }
            }
        }
       else
	 {
           if (debug)
	     {
	      printf("Module has too much space between bars = %d \n", wwidth);
             }
           return(-1);
         }  
      
      bwwidth = bwidth + wwidth;
      tot_width += bwwidth;
      if (debug)
	{
      printf("tot_width = %d bwidth = %d wwidth = %d \n", tot_width,
	     bwidth, wwidth);
        }
    }

  return(tot_width);
}

//
// analyze module   use the ratio of bar widths to get the bits
//      in each module ( instead of looking at onewidth spacing)
//      the data to analyze is in mod_line_data and mod_line_x arrays

int  analyze_module( int col_in, int row_in )
{
  int debug;
  int ii, jj, kk;
  int bwidth;
  int wwidth;
  int bwwidth;
  int tot_width;

  int bit_int;
  int this_mod_index;
  int bitcode;
  int onewid;

  debug = 0;
  lpixel = 0;
  mod_bar_cnt = 0;
  anal_mod_index = -1;

  if (debug)
    {
   printf("Analyzing bad module, mod_pixel_cnt = %d best_col_width_avg = %d \n"
                           , mod_pixel_cnt, best_col_width_avg);
    }
  onewid = (int) onewf; 

  mod_pixel_cnt = best_col_width_avg;

  if (debug) { printf("best_onewf = %3.3f \n",best_onewf); }

  while( lpixel < mod_pixel_cnt)
    {
      //  skip over space

     while((mod_line_data[lpixel] < line_threshold) && (lpixel < mod_pixel_cnt))
      {
       lpixel += 1;
       barwidth += 1;
      }
     if (debug)
       {
        printf("About to get a bar for module , lpixel = %d \n",lpixel);
       }

     if ( lpixel < mod_pixel_cnt)
      {
	  get_mod_bar(row_in,col_in);
      }
    }

  for ( ii = 0; ii < mod_bar_cnt; ii += 1)
    {
      if (debug)
	{
      printf("Mod bar data ii = %d width =  %d space_width = %d  lpixel = %d \n",ii,
           mod_bar_array[ii].width, mod_bar_array[ii].space_width,
                   mod_bar_array[ii].lpixel);
        }
    }

  tot_width = mod_bar_to_mod_bit( .5);
  
  bit_int = 0;
  if (debug) { printf("tot_width = %d \n",tot_width); }

  if ( tot_width == 17)
    {
     bit_int = bits_to_symbol( mod_bit_array );

     if (debug) { printf("bit_int = %x \n", bit_int); }

     this_mod_index = find_index(bit_int,2 , col_in, row_in);

     anal_mod_index = this_mod_index;

     if ( this_mod_index != -1 )
       {
         if (debug) { printf("This mod index = %d \n", this_mod_index); }

	 bitcode = bit_codes[this_mod_index]-1;
         return(0);
         }
     else
       {
         
	 
	 printf("Unable to find module in bit_codes in analyze_module col = %d row = %d\n", col_in, row_in );         
         
       }
    }
  
  if ( debug)
    {
    printf("Module analyze failed, not 17 units wide or bit code not found\n");
    printf("Col = %d Row = %d \n",col_in, row_in );
    }
      // try again without the + .5

    tot_width = mod_bar_to_mod_bit( .25);

    bit_int = 0;
    if ( tot_width == 17)
      {
        bit_int = bits_to_symbol( mod_bit_array );

        if (debug) { printf("bit_int = %x \n", bit_int); }

        this_mod_index = find_index(bit_int,2 , col_in, row_in);

        anal_mod_index = this_mod_index;

        if ( this_mod_index != -1 )
         {
          if (debug) { printf("This mod index = %d \n", this_mod_index); }

	  bitcode = bit_codes[this_mod_index]-1;
          return(0);
         }
       else
        {
	  if (debug)
	    {
	 printf("Unable to find module in bit_codes in analyze_module col = %d row = %d\n", col_in, row_in );         
         printf("Module analyze failed twice\n");
            }
         return(-1);
	}
      }
    else
      {
        if (debug)
          {
	   printf("Module analyze failed, not 17 units wide \n");
          }
        return(-1);
      }

  return(0);
}

void put_back_out( int rowval, int colval , int codeval)
{

  if ( pb_index < 1000)
    {

      pb_array[pb_index].rowval = rowval;
      pb_array[pb_index].colval = colval;
      pb_array[pb_index].codeval = codeval;
      pb_index += 1;
    }

}

void bit_code_out(int rowv, int colv, int bitcode, int val )
{
  int debug;

  debug = 0;
  if ( debug) 
    {
     fprintf( wr_data_file,"%d %d %d %d\n", rowv, colv, bitcode,val);
    }
  else
    {
     fprintf( wr_data_file,"%d %d %d \n", rowv, colv, bitcode);
    }
 
  put_back_out( rowv, colv, bitcode);
}


//
// use a simple ruler to find the module code, slide if necessary
//

int slide_ruler(int testy, int rowin, int indexin, int onewid)
{

  int testx;

  float base;
  int ii;
  int debug;
  int bit_int;
  int bitcode;
  int rowval;
  int this_mod_index;
  int bit[100];

  debug = 0;
  testx = (onewid /2) ;
  base = best_onewf / 2;

  this_mod_index = -1;

  if (( onewid /4 ) > 0)
     {
      base = best_onewf * .2;
     
       while(( this_mod_index == -1) && ( base < (best_onewf * 1.3)))
	 {
          for (ii=0; ii < 17 ; ii += 1)
           {
            testx = (int) ( base + (( float)ii * best_onewf));
            if (debug)
             {
              printf("In slide_ruler, testx , testy = %d %d  ii = %d \n", 
                          testx, testy, ii);
             }
            bit[ii] = test_black( testx, testy, 1);
	   }     
          if (debug)
           {
            for (ii = 0; ii < 17; ii += 1)
             {
              printf("ii = %d bit[ii] = %d \n", ii, bit[ii]); 
             }
           }
   
         bit_int = bits_to_symbol( bit );

         if (debug) { printf("bit_int = %x \n", bit_int); }

         this_mod_index = find_index(bit_int,2 , indexin+1, rowin+1);

         if ( this_mod_index != -1 )
          {
            if (debug) { printf("This mod index = %d \n", this_mod_index); }
   
            slide_mod_index = this_mod_index;

            bitcode = bit_codes[this_mod_index]-1;
            rowval = bitcode / 929;
            if ( rowval != (rowin % 3))
             {
	      printf("Error:  This bar code is not a row %d code \n",rowin+1);
             }
            else
             {
              if (debug) 
	       {
	       printf("row = %d col = %d bitcode = %d \n", rowin+1, indexin+1, 
                     bitcode );
               }
	      //             bit_code_out( rowin, indexin, bitcode);
              return(0);                         
             }        
           }
         else
          {
          if (debug)
           {
            printf("The module unshifted code is incorrect - column = %d row = %d module = 0x%x\n",
	       indexin+1, rowin+1, bit_int); 
           }   
         }

       base = base + ( .1 * best_onewf);
      }
     }
  return(-1);
}

//
// use a simple ruler to find the module code
//

int check_ruler(int testy, int rowin, int indexin)
{

  int testx;
  float base;
  int ii;
  int debug;
  int bit_int;
  int bitcode;
  int rowval;
  int this_mod_index;
  int bit[100];
  debug = 0;
  
  base = best_onewf / 2;

  for (ii=0; ii < 17 ; ii += 1)
   {
    testx = (int) ( base + (( float)ii * best_onewf));
    if (debug)
      {
       printf("In check_ruler, testx , testy = %d %d  ii = %d \n", 
                          testx, testy, ii);
      }
    bit[ii] = test_black( testx, testy, 1);
             
    }

  if (debug)
    {
     for (ii = 0; ii < 17; ii += 1)
       {
         printf("ii = %d bit[ii] = %d \n", ii, bit[ii]); 
       }
    }
   
   bit_int = bits_to_symbol( bit );

   if (debug) { printf("bit_int = %x \n", bit_int); }

   this_mod_index = find_index(bit_int,2 , indexin+1, rowin+1);

   rule_mod_index = -1;

   if ( this_mod_index != -1 )
     {
      if (debug) { printf("This mod index = %d \n", this_mod_index); }

      rule_mod_index = this_mod_index;
      bitcode = bit_codes[this_mod_index]-1;
      rowval = bitcode / 929;
      if ( rowval != (rowin % 3))
        {
	  printf("Error:  This bar code is not a row %d code \n",rowin+1);
          return(-1);
        }
      else
	{
         if (debug) 
	  {
	    printf("row = %d col = %d bitcode = %d \n", rowin+1, indexin+1, 
                       bitcode );
          }
         return(0);      
        }        
      }
    else
     {
      if (debug)
        {
        printf("The module unshifted code is incorrect - column = %d row = %d module = 0x%x\n",
	       indexin+1, rowin+1, bit_int); 
        }   
      return(-1);
     }

}

//
// get a column
//
void get_a_column(int indexin, int startx, int onewid)
{
  int ii;
  int jj;
  int kk;
  int nn;
  int ll,mm;

  float base;
  int col_width;
  int col_height;
  char namestr[120];
  int debug;
  int testx;
  int testy;
  int this_mod_index;
  int bitcode;
  int bit_int;
  int bit[30];
  int rowval;
  int yinc;
  char barfilestr[120];
  int anal_error;
  int rule_error;
  int slide_error;
  int xline_err;
  int agree_flag;

  debug = 0;

  if ( debug ) 
    {
     printf("In get a column, pixel_out_ylen = %d \n",best_pixel_out_ylen);
     printf("indexin = %d  startx = %d onewid = %d \n",
            indexin, startx, onewid );
    }

  sprintf(namestr,"outcol%d.pgm",indexin+1);

  onewid = (int) best_onewf; 
  nn = 0;
  for (ii = 0; ii < best_pixel_out_ylen; ii += 1)
    {
       if ( pix_final[(3 * onewid)][ii].greyval > 0)
        {
          
	  kk = 0;
          for ( jj = startx; jj < ( startx + best_begin_left_col + onewid );
                               jj += 1)
           {
	     if (( kk < OUT_MAX_X_PIXEL) && ( nn < OUT_MAX_Y_PIXEL))
	       {
	        pix_cols[kk][nn].greyval = pix_final[jj][ii].greyval;
	        kk += 1;
               }
             else
	       {
		 printf("pix_cols array size exceeded \n");
                 exit(0);
               }
           }
	nn += 1;
	col_width = kk;
       	}
       
    }

  col_height = nn;

  if (debug)
   {
    printf("About to call write_pgm_outcol - namestr = %s \n", namestr);
   }

  if ( write_cols)
    {
      //  write_pgm_outcol(namestr, (20 * onewidth), nn);
      write_pgm_outcol( namestr, best_col_width_avg, nn);
    }

  // now get the row data


  this_mod_index = -1;
  slide_mod_index = -1;
  rule_mod_index = -1;
  xline_mod_index = -1;

  do_pixcol_diffy(col_width, col_height);
  
  testy = (best_row_height / 2);
  yinc = best_row_height;

  if (debug) { printf("best_row_height = %d \n", best_row_height); }

  kk = 0;
  while( testy < col_height)
    {
          mm = 0;
	  for ( ll = (testy - (yinc/2)); ll < (testy + (yinc/2)); ll += 1)
	  {
           for( ii = 0; ii < best_col_width_avg; ii += 1)
	    {
              if ( ( ii < MAX_XPIXEL_MODULE   ) && ( mm < MAX_YPIXEL_MODULE))
		{
                 pix_module[ii][mm].greyval  =  pix_cols[ii][ll].greyval;
                }
              else
                {
                  printf("Size of pix_module = %d %d \n", MAX_XPIXEL_MODULE,
                            MAX_YPIXEL_MODULE);
                }
            }
           mm += 1;
	  }
          
	  if ( write_mods )
	    {       
	     sprintf(barfilestr,"outmodule%d_%d.pgm",kk+1,indexin+1);
              // write_pgm_outmod( barfilestr, (17 * onewid) , mm);
	      write_pgm_outmod( barfilestr, best_col_width_avg, mm);
            }

	  // copy the module into mod_line, needed by analyze module

           for( ii = 0; ii < (best_col_width_avg); ii += 1)
            {
	     mod_line_data[ii] = pix_cols[ii][testy].greyval;
             mod_line_x[ii] = ii;
            }

	   //  mod_pixel_cnt = ii;    
	   mod_pixel_cnt = ( int) ( 17.0 * best_onewf);

           anal_error =analyze_module(indexin+1, kk+1); 
                                     // analyze module according to 
                                     // bar ratios
          if ( anal_error == 0)    // able to analyze
     	    {          

	      xline_err = get_mod_xline( testy, indexin+1, kk+1);
              if ( xline_err == 0)
                {
                  if ( xline_mod_index != anal_mod_index)
                    {
                      if (debug)
			{
		      printf("Xline and analyze disagree: ");
                      printf("xline_mod_index = %d anal_mod_index = %d \n",
			     xline_mod_index, anal_mod_index );
                        }
                    }
		}
              else
                {
		  if (debug) { printf("Xline failed \n"); }
                }

             this_mod_index = anal_mod_index;
             bitcode = bit_codes[this_mod_index]-1;
             rowval = bitcode / 929;
             if ( rowval != (kk % 3))
               {
                 printf("Error:  This module is not a row %d code \n",kk+1);
               }
             else
               {
                if ( debug) 
	       	 {
                   printf("row = %d col = %d bitcode = %d \n", kk+1,
                              indexin+1, bitcode);
                 }			    
                bit_code_out(kk, indexin, bitcode , 0);
       	       }
             }
	  else      // bitcode not found by analyze module
             {
             if (debug)
	     	{
                 printf("Analyze module failed \n");
                }

	      xline_err = get_mod_xline( testy, indexin+1, kk+1);
              if ( xline_err == 0)
                {
                 this_mod_index = xline_mod_index;
                 bitcode = bit_codes[this_mod_index]-1;
                 rowval = bitcode / 929;
                 if (debug)
		   {
                    printf("Analyze failed - but xline success: ");
                    printf("xline_mod_index = %d,row=%d col=%d bitcode = %d\n",
                     xline_mod_index,kk+1,indexin+1,bitcode);
                   }
		 bit_code_out(kk, indexin, bitcode,1);
		}

             rule_error = check_ruler(testy, kk, indexin );

             if ( rule_error == 0)
               {
		 this_mod_index = rule_mod_index;
                 agree_flag = FALSE;
                 if ( xline_err == 0)
                   {
                    if ( rule_mod_index != xline_mod_index)
		     {
		       if (debug)
			 {
		         printf("rule index and xline index do not agree: ");
                          printf("rule_mod_index = %d row = %d col = %d\n", 
                             rule_mod_index,kk+1,indexin+1);
                         }
		      // this_mod_index = xline_mod_index;
		      //     bitcode = bit_codes[this_mod_index]-1;
                      rowval = bitcode / 929;

                     }
                   else
		    {
                      agree_flag = TRUE;
                      if (debug)
			{
		         printf("rule index and xline index agree \n");
                        }
                    }
                   }
                  bitcode = bit_codes[this_mod_index]-1;
                  rowval = bitcode / 929;
                  if (agree_flag == FALSE)  
                    {
		      if ( xline_err != 0)  // if we haven't put out
                        {                      // because xline failed
                         bit_code_out(kk, indexin, bitcode,2 );
                        }
                    }
                }
	     else
               {
                 if (debug)
		   {
                    printf("Both analyze module and check_ruler failed \n");
                   }
                 slide_error = slide_ruler(testy, kk, indexin, onewid );

                 if ( slide_error == 0)
                   {
		     this_mod_index = slide_mod_index;
                     agree_flag = FALSE;
                     if ( xline_err == 0)
		       {
                         if (slide_mod_index != xline_mod_index)
                           {
                             if (debug)
			      {
			    printf("Slide index and xline index disagree \n");
                              }
			     //  this_mod_index = xline_mod_index;
                             bitcode = bit_codes[this_mod_index]-1;
                             rowval = bitcode / 929;
			     //   bit_code_out(kk, indexin, bitcode,1);
                           }
                         else
                           {
                             if (debug) {
			       printf("Slide index and xline index agree \n"); }
                             agree_flag = TRUE;
                           }
                        }
                     bitcode = bit_codes[this_mod_index]-1;
                     rowval = bitcode / 929;
                     if ( agree_flag == FALSE)
		       {
			 if ( slide_mod_index != rule_mod_index)
                           {
			     if ( xline_err != 0) 
                               {
                                bit_code_out(kk, indexin, bitcode,3);
                               }
                           }
                       }
                   }
		 else  // slide error fails
                   {
                     if ( debug)
		       {
		     printf("All analyze of module failed: row= %d col= %d\n",
                                  kk+1, indexin+1);
                       }
		     if ( xline_err != 0)
		       {
			bit_code_out(kk, indexin, 9999, 4 );
                       }
                     else
                       {
			 this_mod_index = xline_mod_index;
                        bitcode = bit_codes[this_mod_index]-1;
                        rowval = bitcode / 929;
			//    bit_code_out(kk, indexin, bitcode,5);
                       }
                   }
               }
             }
    
         if ( ( kk < MAX_SYMBOL_ROWS) && ( indexin < MAX_SYMBOL_COLS))
         {
          symbol_data[kk][indexin] = this_mod_index;
         }
        else
         {
          if ( kk > MAX_SYMBOL_ROWS)
           {
            printf("Too many rows in the symbol \n");
           }
          if ( indexin > MAX_SYMBOL_COLS)
           {
            printf("Too many columns in the symbol \n");
           }
         }
        if ( debug)
         {
          printf("row,col = %d %d this_mod_index = %d \n", kk+1,indexin+1,
                                this_mod_index);
         }

     testy = testy + yinc;
     kk += 1;
    }  // for each row
  
}

//
// get the whole left column of the symbol from pix_final
//
void get_pixel_left_col()
{

  int ii;
  int jj;
  int kk;
  int nn;
  int left_col_width;
  int left_col_height;
  int onewid;
  char namestr[120];

  int debug;
  debug = 0;

  if (debug) { printf("best_pixel_out_xlen = %d \n",
                      best_pixel_out_xlen); }

  if (debug) { printf("best_pixel_out_ylen = %d \n",
                      best_pixel_out_ylen); }

  if (debug) {
            printf("In get_pixel_left_col Onewidthf  = %f best_onewf = %f\n",
		      onewidthf , best_onewf); }

   write_pgm_outpix_final( best_pixel_out_xlen, best_pixel_out_ylen );

  if ( debug ) 
    { printf("In get pixel left col, pixel_out_ylen = %d \n",pixel_out_ylen); }

  onewid = (int) best_onewf; 

  if ( debug) { printf("In get pixel left col, onewid = %d \n", onewid); }

  nn = 0;
  for (ii = 0; ii < best_pixel_out_ylen; ii += 1)
    {
       
       if ( pix_final[(3 * onewid)][ii].greyval > 0)
        {
          if ( debug) { printf("ii = %d is grey \n",ii ); }   
	  kk = 0;
          for ( jj = best_begin_left_col; jj < ( best_begin_left_col + (20*onewid));
                               jj += 1)
           {
	    pix_cols[kk][nn].greyval = pix_final[jj][ii].greyval;
	    kk += 1;
           }
	nn += 1;
	left_col_width = kk;
       	}
       
    }

  left_col_height = nn;
  if (debug) { printf("Calling do_pixcol_diffx \n"); }

  do_pixcol_diffx( left_col_width, left_col_height );    
              // Find 1st derivative in x direction
  if (debug) { printf("Calling do_pixcol_diffy \n"); }

  do_pixcol_diffy( left_col_width, left_col_height );    
              // Find 1st derivative in y direction
  if (debug) { printf("Calling do_top_bar \n"); }

  do_top_bar(left_col_height);


  if ( best_col_width_avg == 0 )
    {
      best_col_width_avg = 17 * onewidth;
    }

  if (debug) 
    {
     printf("About to call write_pgm_outcal from get_pixel_left_col\n"); 
     printf("best_col_width_avg = %d \n", best_col_width_avg); }
 
  if ( write_leftcol == TRUE )
    {
      //     write_pgm_outcol( "outcol_left.pgm", (20 * onewidth), nn);
      write_pgm_outcol( "outcol_left.pgm", best_col_width_avg, nn);
    }

  
}

// having found the best qualifying unslanted symbol
// process it
//
void do_best(int xlenin, int ylenin)
{
  int debug;

  debug = 0;

  if ( debug )
    {
      printf("In do best , xlenin = %d ylenin = %d \n", xlenin, ylenin);
    }

  //   find_left_edge( best_normal_x, best_normal_y, best_normal_slope,
  //            best_normal_dir, b_box_ylen/2 , b_box_xlen);

   get_pixel_left_col();    // get the leftmost column of data

   get_columns_start( ylenin);

   //get the leftmost row data from the pixel_final array

   if (write_final_pix )
     {     
      write_pgm_outpix_final( xlenin, ylenin );
      }

}

// find height of symbol

void find_height(int nx, int ny, int nslope, int ndir, int ylen)
{
  int ii, jj;
  int xinc, yinc;
  int dark_percent;

  int bits_on_cnt;
  int num_dark;
  int first_dark;
  int first_dark_ind;
  int first_light;
  int first_light_ind;
  int debug;
  
  debug = 0;

  if (debug) 
    {
     printf("In find height , x,y = %d %d \n",nx, ny );
     printf("Slope = %d \n", nslope);
     printf("ylen = %d \n", ylen);
     printf("dir = %d \n",ndir);
    }


  if ( ndir == 1)  // positive x and positive y incr
    {

       yinc = (ylen);
       xinc = (((ylen) * nslope) / 100);
       if (debug)
	 {
          printf("In find height - doing line from  x,y = %d %d  to \n",
	      nx-xinc, ny-yinc );
          printf("to x,y = %d %d \n", nx+xinc, ny+yinc);
         }

      line_fast_norm(nx-xinc, ny-yinc, nx + xinc, ny+yinc);
    }

  if ( ndir == -1)   // negative x incr and positive y incr
    {
      
       yinc = (ylen);
       xinc = 0- (( (ylen) * nslope) / 100);
       if (debug)
	 {
          printf("In find height - doing line from  x,y = %d %d  to \n",
	      nx-xinc, ny-yinc );
          printf("to x,y = %d %d \n", nx+xinc, ny+yinc);
         }

      line_fast_norm(nx-xinc, ny-yinc, nx + xinc, ny+yinc);
    }


  first_dark = FALSE;
  first_dark_ind = -1;
  ii = 0;
  while ( (ii < pixel_cnt) && ( first_dark == FALSE))
    {
      if ( norm_line_data[ii] > dark_threshold)
	{
          if ( first_dark == FALSE)
	    {
	      first_dark = TRUE;
              first_dark_ind = ii;
            }
        }
      ii += 1;
    }

  if ( first_dark == TRUE)
    {
    first_light = FALSE;

    while ( (ii < pixel_cnt) && ( first_light == FALSE))
     {
      if ( norm_line_data[ii] < dark_threshold)
	{
          if ( first_light == FALSE)
	    {
	      first_light = TRUE;
              first_light_ind = ii;
            }
        }
      ii += 1;
    }
    if ( first_light == TRUE )
      {
	symbol_height = first_light_ind - first_dark_ind;
      }
   }
 
  if (debug)
    {
     printf("Symbol height = %d \n",symbol_height);
    }
  
}

// find left edge of symbol
//    in pixel_array data , then unslant it 
//    and put the result pix_out
//

void find_left_edge(int nx, int ny, int nslope, int ndir, int ylen, int xlen)
{
  int ii, jj;
  int xinc, yinc;
  int dark_percent;

  int bits_on_cnt;
  int num_dark;
  int first_dark;
  int first_dark_ind;
  int first_light;
  int first_light_ind;
  int save_pixel_cnt;
  int debug;

  debug = 0;

  if (debug)
    {
     printf("In find left edge , x,y = %d %d \n",nx, ny );
     printf("Slope = %d \n", nslope);
     printf("ylen = %d \n", ylen);
     printf("xlen = %d \n", xlen);
     printf("dir = %d \n",ndir);
    }

  if ( ndir == 1)  // positive x and positive y incr
    {

       yinc = (ylen);
       xinc = (((ylen) * nslope) / 100);
       if (debug)
	 {
          printf("In find left edge  - doing line from  x,y = %d %d  to \n",
	      nx-xinc, ny-yinc );
          printf("to x,y = %d %d \n", nx+xinc, ny+yinc);
	 }
      line_fast_norm(nx-xinc, ny-yinc, nx + xinc, ny+yinc);
    }

  if ( ndir == -1)   // negative x incr and positive y incr
    {
      
       yinc = (ylen);
       xinc = 0- (( (ylen) * nslope) / 100);
       if (debug)
	 {
          printf("In find left edge - doing line from  x,y = %d %d  to \n",
	      nx-xinc, ny-yinc );
          printf("to x,y = %d %d \n", nx+xinc, ny+yinc);
         }

      line_fast_norm(nx-xinc, ny-yinc, nx + xinc, ny+yinc);
    }


  // use each x,y point in the normal data as a starting point for
  // a new line of slope = current line slope and length equal to width
  // of bounding box 

  save_pixel_cnt = pixel_cnt;
  pixel_out_ylen = save_pixel_cnt;

  pixel_out_xlen = -1;

  ii = 0;
  while  (ii < save_pixel_cnt) 
    {
 
     if ( ndir == 1)  // That is the dir of the original = -1
       {

         xinc = (xlen);
         yinc = 0- (( (xlen) * nslope) / 100);

         if (debug)
	   {
            printf("In find left edge  - doing line from  x,y = %d %d  to \n",
	      normal_x[ii], normal_y[ii] );

            printf("to x,y = %d %d \n", normal_x[ii] + xinc,
                                normal_y[ii] + yinc);

            printf(" xinc yinc = %d %d \n",xinc, yinc);
           }

         line_fast(normal_x[ii], normal_y[ii], normal_x[ii] + xinc,
			normal_y[ii] + yinc);
       }
  

   if ( ndir == -1)   // That is the dir of the original = 1
      {
         xinc = (xlen);
         yinc = (((xlen) * nslope) / 100);
         if (debug)
	   {
            printf("In find left edge  - doing line from  x,y = %d %d  to \n",
	      normal_x[ii], normal_y[ii] );
            printf("to x,y = %d %d \n", normal_x[ii] + xinc, 
                                       normal_y[ii] + yinc );

            printf(" xinc yinc = %d %d \n",xinc, yinc);
           }
    
          line_fast(normal_x[ii], normal_y[ii], normal_x[ii] + xinc,
		 normal_y[ii]  + yinc);
    }

   // copy the line of data just found to the output

   if ( pixel_cnt > pixel_out_xlen)
     {
       pixel_out_xlen = pixel_cnt;
     }

   for (jj = 0; jj < pixel_cnt; jj += 1)
     {
       pix_out[jj][ii].greyval = line_data[jj];
     }

   ii += 1;
  }

  if (write_outpix)
    {
     write_pgm_outpix( pixel_cnt, save_pixel_cnt );
    }

  if (debug)
    {
      printf("End find left edge \n");
    }

}


// post process pix_out
//    by correcting for the pix_slope value
//      looks at pix_slope
void final_pix(int xlen, int ylen )
{
  int ii, jj;
  int xinc, yinc;
  int debug;

  debug = 0;

  if (debug)
    {
     printf("In final pix , xlen = %d ylen = %d \n", xlen, ylen);

     printf("pix_slope = %d \n", pix_slope);
    }
 
  for ( ii = 0; ii < ylen  ; ii += 1)
    {
     if ( pix_slope > 0 )  // positive x and positive y incr
      {
       xinc = xlen;
       yinc = (((xlen) * pix_slope) / 1000);
       
       line_fast_post(0, ii , xinc, ii + yinc);
      }

     if ( pix_slope < 0 )   // negative x incr and positive y incr
      {
      
       xinc = xlen;
       yinc = ( xlen * pix_slope) / 1000;
       
       line_fast_post(0, ii, xinc, ii + yinc);
      }
     
     if ( pix_slope == 0)
       {
	 xinc = xlen;
         yinc = 0;
         line_fast_post(0, ii, xinc, ii );
       }
   
     for(jj=0; jj < post_pixel_cnt; jj += 1)
       {
         if (( ii < OUT_MAX_Y_PIXEL) && ( jj < OUT_MAX_X_PIXEL ))
	   {
	    pix_final[jj][ii].greyval = post_line_data[jj];
           }
         else
           {
	     printf("Barcode symbol too big for pix_final , try shrinking\n");
             exit(0);
           }
       }
  }

  if ( pix_slope != 0 )
    { 
     unslant_pix_final( xlen, ylen);
    }

  //get_pixel_left_col();   get the leftmost column of data

  //  get_columns_start(xlen, ylen);

  // get the leftmost row data from the pixel_final array

  //  write_pgm_outpix_final( xlen, ylen );

}

void get_scan_lines(int inx, int iny, int inflag)
{
  int kk;
  int xstart;
  int ystart;
  int yinc;
  int xinc;
  int debug;

  debug = 0;

  // do lower left case

  if (inflag == 2)
    {
     for (kk = 0; kk < 20; kk += 1)
      {
	dir = 1;
       slope = slope_array[kk];
       xstart =  inx;
       ystart =  iny;

       xinc = b_box_xlen;
       yinc = (( b_box_xlen * slope) / 100);
       
       pixel_cnt =0;
       

       line_fast( xstart, ystart, xstart + xinc, ystart + yinc);

       if (debug)
	 {       
          printf("Did ll line at start = %d %d to fin = %d %d \n",
	      xstart, ystart, xstart + xinc, ystart + yinc );
         }
       smooth_line();
      
       save_line();

       decode_line();
       

      }
    }

  // do center case

  if (inflag == 1)
    {
     for (kk = 0; kk < 10; kk += 1)
      {
	dir = 1;
       slope = slope_array[kk];
       xstart =  inx;
       ystart =  iny;
 
       xinc = b_box_xlen;
       yinc = (( b_box_xlen * slope) / 100);

       pixel_cnt = 0;       

       line_fast( xstart, ystart, xstart + xinc, ystart + yinc);
       
       if (debug)
	 {
         printf("Did cl+ line at start = %d %d to fin = %d %d \n",
	      xstart, ystart, xstart + xinc, ystart + yinc );
         }

       smooth_line();
       save_line();


       decode_line();


      }
     for (kk = 0; kk < 10; kk += 1)
      {
	dir = -1;
       slope = slope_array[kk];
       xstart =  inx;
       ystart =  iny;

       xinc = b_box_xlen;
       yinc = 0 - (( b_box_xlen * slope) / 100);

       pixel_cnt = 0;       


       line_fast( xstart, ystart, xstart + xinc, ystart + yinc);

       if (debug)
	 {       
          printf("Did cl- line at start = %d %d to fin = %d %d \n",
	      xstart, ystart, xstart + xinc, ystart + yinc );
         }
       smooth_line();
       save_line();


       decode_line();


      }
    }

  if (inflag == 0)      // upper left case
    {
     for (kk = 0; kk < 20 ; kk += 1)
      {
	dir = -1;
       slope = slope_array[kk];
       xstart =  inx;
       ystart =  iny;

       xinc = b_box_xlen;
       yinc = 0- (( b_box_xlen * slope) / 100);

       pixel_cnt = 0;       

       line_fast( xstart, ystart, xstart + xinc, ystart + yinc);
       
       if (debug)
	 {
          printf("Did ul line at start = %d %d to fin = %d %d \n",
	      xstart, ystart, xstart + xinc, ystart + yinc );
         }

       smooth_line();
       save_line();


       decode_line();


      }
    }

}

void write_pgm_file( )
{
  FILE *wrfile;

  int numcols;
  int numrows;
  int maxval;

  int forceval;

  char pgm_ofile_str[120];

  forceval = 0;

  strcpy(pgm_ofile_str,"out.pgm");

  wrfile = fopen(pgm_ofile_str,"wb");

  maxval = 255;

  if ( wrfile == NULL)
    {
      printf("Unable to open the pbm output file %s\n",pgm_ofile_str);
      exit(-1);
    }
 
 
  pgm_writepgm (wrfile, greydata , numcols , numrows , (gray) maxval, forceval);

}

int read_pgm_in()
{
  
  char str[30];
  FILE *fp;
  int grayv;
  int row_size;
  int col_size;
  int ii,jj;
  char linedata[1000];
  int tint;
  int debug;
  size_t tsize;
  int kk;
  char *vptr;
  char *tptr;
  int good_format;

  debug = 0;

  if (debug) { printf("In read_pgm_in \n"); }

  // strcpy(pgm_file_str,"in.pgm");
 

  if((fp=fopen(pgm_infile_str,"rb")) == NULL)
    {

      printf("Error opening input file  %s  \n", pgm_infile_str);
      exit(-1);
    }
  else

   {
    fscanf(fp, "%s\n", str);
    
    good_format = FALSE;

    if (strcmp (str, "P5") == 0)

      {
        printf( "\n Image file %s in PGM format \n", pgm_infile_str);
        good_format = TRUE;
        bytes_per_pix = 1;        
      }

    if (strcmp (str, "P6") == 0)

      {
        printf( "\n Image file %s in PNM color format \n", pgm_infile_str);
        good_format = TRUE;
        bytes_per_pix = 3;
        printf("Must have grey scale format - save image in GIMP as greyscale\n");
        exit(0);
      }

    if ( good_format == FALSE)
      {
	printf("\n Image file = %s is not in PGM or PNM binary format \n",
	       pgm_infile_str);
        exit(0);
      }

/***************************************************************************
* Read header information
***************************************************************************/

      do{ fgets(linedata,9999,fp);

         } while(linedata[0] == '#');
   
      sscanf(linedata,"%d %d",&col_size,&row_size);

      fscanf(fp, "%d\n", &grayv) ;
   }

printf("col_size = %d, row_size = %d, gray levels are = %d \n", 
             col_size, row_size,grayv);

if ((row_size < 0) || (col_size < 0))

{
   printf("\n stmt 2 :Image file %s not in PGM or PNM greyscale format\n",
                  pgm_infile_str);
   printf("%s\n",linedata);
   return(0);
}

  number_of_cols = col_size;
  number_of_rows = row_size;

  if ( number_of_cols > MAX_X_PIXEL)
    {
      printf("Image too big in x direction \n");
    }
  if ( number_of_rows > MAX_Y_PIXEL )
    {
      printf("Image too big in y direction \n");
    }

  vptr = malloc((size_t) (col_size * row_size  * bytes_per_pix));
 
  if ( vptr != NULL)
    {
     tsize = fread(vptr, (size_t) 1, 
                 (size_t) ( bytes_per_pix *col_size * row_size), fp);
     
    }
  else
    {
      printf("Failed malloc \n");
    }

  debug = 0;

  if (bytes_per_pix == 1)
    {
     for(jj=0;jj<row_size;jj++)
      {
       for(ii=0;ii<col_size;ii++)
        {
         tint = ((int) *vptr) & 255;
         if (debug) {
	  printf("ii = %d jj = %d kk = %d tint = %x \n", ii,jj,kk,tint); }
          vptr++;
         if ( tint > 127 )
          {
           pixel_array[ii][jj].greyval =  256 - tint  - 1;
          }
         else
          {
           pixel_array[ii][jj].greyval =  128 + tint;
          }
        }
      }
    }


 free(vptr);
fclose(fp);


 if ( debug) { printf("Completed read of input file \n"); }

 return(1);
}

void write_pgm_out( int num_cols, int num_rows)
{
  char pgm_file_str[120];
  char str[30];
  FILE *fp;
  int grayv;
  int row_size;
  int col_size;
  int ii,jj;
  char linedata[1000];
  int tint;
  int tchar;
  char row_str[40];
  char col_str[40];
  char outstr[80];
  char gray_str[20];

  grayv = 255;
  row_size = num_rows;
  col_size = num_cols;

  strcpy(pgm_file_str,"out.pgm");

  if((fp=fopen(pgm_file_str,"w")) == NULL)

   printf("Error opening %s file \n", pgm_file_str);

  else

   {
     strcpy(str,"P5\n");
     fputs(str,fp);
    
     strcpy(str,"#SANE data follows \n");
     fputs(str,fp);

/***************************************************************************
*Write header information
***************************************************************************/

     sprintf(col_str,"%d ",col_size);
     sprintf(row_str,"%d\n",row_size);
     strcpy(outstr,col_str);
     strcat(outstr,row_str);

     fputs(outstr,fp);
     
     strcpy(gray_str,"255\n");
     fputs(gray_str,fp);
   }




for(jj=0;jj<row_size;jj++)
 {
   for(ii=0;ii<col_size;ii++)
     {
       tint = pixel_array[ii][jj].greyval;
    
       if ( tint > 127)
         {
           tint = tint - 128;
         }
       else
         {
           tint = (255 - tint) | 0x80;
         }
       tchar = (char) tint;
       fputc(tchar,fp);
     }
  }

fclose(fp);

}

void write_pgm_outcol(char *namestrin,  int num_cols, int num_rows)
{
  char pgm_file_str[120];
  char str[30];
  FILE *fp;
  int grayv;
  int row_size;
  int col_size;
  int ii,jj;
  char linedata[1000];
  int tint;
  int tchar;
  char row_str[40];
  char col_str[40];
  char outstr[80];
  char gray_str[20];

  grayv = 255;
  row_size = num_rows;
  col_size = num_cols;

  strcpy(pgm_file_str,namestrin);

  if((fp=fopen(pgm_file_str,"w")) == NULL)

   printf("Error opening %s file \n", pgm_file_str);

  else

   {
     strcpy(str,"P5\n");
     fputs(str,fp);
    
     strcpy(str,"#SANE data follows \n");
     fputs(str,fp);

/***************************************************************************
*Write header information
***************************************************************************/

     sprintf(col_str,"%d ",col_size);
     sprintf(row_str,"%d\n",row_size);
     strcpy(outstr,col_str);
     strcat(outstr,row_str);

     fputs(outstr,fp);
     
     strcpy(gray_str,"255\n");
     fputs(gray_str,fp);
   }


for(jj=0;jj<row_size;jj++)
 {
   for(ii=0;ii<col_size;ii++)
     {
       tint = pix_cols[ii][jj].greyval;
       if ( tint > 127)
         {
           tint = tint - 128;
         }
       else
         {
           tint = (255 - tint) | 0x80;
         }
       tchar = (char) tint;
       fputc(tchar,fp);
     }
  }

fclose(fp);

}


// write out the single modules
//   in pgm format
//
void write_pgm_outmod(char *namestrin,  int num_cols, int num_rows)
{
  char pgm_file_str[120];
  char str[30];
  FILE *fp;
  int grayv;
  int row_size;
  int col_size;
  int ii,jj;
  char linedata[1000];
  int tint;
  int tchar;
  char row_str[40];
  char col_str[40];
  char outstr[80];
  char gray_str[20];

  grayv = 255;
  row_size = num_rows;
  col_size = num_cols;

  strcpy(pgm_file_str,namestrin);

  if((fp=fopen(pgm_file_str,"w")) == NULL)

   printf("Error opening %s file \n", pgm_file_str);

  else

   {
     strcpy(str,"P5\n");
     fputs(str,fp);
    
     strcpy(str,"#SANE data follows \n");
     fputs(str,fp);

/***************************************************************************
*Write header information
***************************************************************************/

     sprintf(col_str,"%d ",col_size);
     sprintf(row_str,"%d\n",row_size);
     strcpy(outstr,col_str);
     strcat(outstr,row_str);

     fputs(outstr,fp);
     
     strcpy(gray_str,"255\n");
     fputs(gray_str,fp);
   }


for(jj=0;jj<row_size;jj++)
 {
   for(ii=0;ii<col_size;ii++)
     {
       tint = pix_module[ii][jj].greyval;
       if ( tint > 127)
         {
           tint = tint - 128;
         }
       else
         {
           tint = (255 - tint) | 0x80;
         }
       tchar = (char) tint;
       fputc(tchar,fp);
     }
  }

fclose(fp);

}

void write_pgm_outpix( int num_cols, int num_rows)
{
  char pgm_file_str[120];
  char str[30];
  FILE *fp;
  int grayv;
  int row_size;
  int col_size;
  int ii,jj;
  char linedata[1000];
  int tint;
  int tchar;
  char row_str[40];
  char col_str[40];
  char outstr[80];
  char gray_str[20];

  grayv = 255;
  row_size = num_rows;
  col_size = num_cols;

  strcpy(pgm_file_str,"out.pgm");

  if((fp=fopen(pgm_file_str,"w")) == NULL)

   printf("Error opening %s file \n", pgm_file_str);

  else

   {
     strcpy(str,"P5\n");
     fputs(str,fp);
    
     strcpy(str,"#SANE data follows \n");
     fputs(str,fp);

/***************************************************************************
*Write header information
***************************************************************************/

     sprintf(col_str,"%d ",col_size);
     sprintf(row_str,"%d\n",row_size);
     strcpy(outstr,col_str);
     strcat(outstr,row_str);

     fputs(outstr,fp);
     
     strcpy(gray_str,"255\n");
     fputs(gray_str,fp);
   }

for(jj=0;jj<row_size;jj++)
 {
   for(ii=0;ii<col_size;ii++)
     {
       tint = pix_out[ii][jj].greyval;
    
       if ( tint > 127)
         {
           tint = tint - 128;
         }
       else
         {
           tint = (255 - tint) | 0x80;
         }
       tchar = (char) tint;
       fputc(tchar,fp);
     }
  }

fclose(fp);
 
}

void  write_pgm_outpix_final( int num_cols, int num_rows)
{
  char pgm_file_str[120];
  char str[30];
  FILE *fp;
  int grayv;
  int row_size;
  int col_size;
  int ii,jj;
  char linedata[1000];
  int tint;
  int tchar;
  char row_str[40];
  char col_str[40];
  char outstr[80];
  char gray_str[20];

  grayv = 255;
  row_size = num_rows;
  col_size = num_cols;

  strcpy(pgm_file_str,"outf.pgm");

  if((fp=fopen(pgm_file_str,"w")) == NULL)

   printf("Error opening %s file \n", pgm_file_str);

  else

   {
     strcpy(str,"P5\n");
     fputs(str,fp);
    
     strcpy(str,"#SANE data follows \n");
     fputs(str,fp);

/***************************************************************************
*Write header information
***************************************************************************/

     sprintf(col_str,"%d ",col_size);
     sprintf(row_str,"%d\n",row_size);
     strcpy(outstr,col_str);
     strcat(outstr,row_str);

     fputs(outstr,fp);
     
     strcpy(gray_str,"255\n");
     fputs(gray_str,fp);
   }

for(jj=0;jj<row_size;jj++)
 {
   for(ii=0;ii<col_size;ii++)
     {
       tint = pix_final[ii][jj].greyval;
    
       if ( tint > 127)
         {
           tint = tint - 128;
         }
       else
         {
           tint = (255 - tint) | 0x80;
         }
       tchar = (char) tint;
       fputc(tchar,fp);
     }
  }

fclose(fp);

}

void read_pgm_file( )
{
  FILE *rdfile;
  int ii, jj;

  int numcols;
  int numrows;
  int maxval;

  char pgm_file_str[120];

  printf("About to read the pgm file \n");

  strcpy(pgm_file_str,"in.pgm");

  printf("pgm_file_str = %s \n",pgm_file_str);

  rdfile = fopen(pgm_file_str,"rb");

  if ( rdfile == NULL)
    {
      printf("Unable to open the pgm file = %s \n",pgm_file_str);
      exit(-1);
    }

  //  greydata = (char *) malloc(1000*sizeof(char));
 
  greydata = pgm_readpgm (rdfile, &numcols , &numrows , &maxval);

  number_of_cols = numcols;
  number_of_rows = numrows;

  printf("Returned from readpgm- number of columns = %d number of rows = %d \n",
	 numcols, numrows );

  if ( number_of_cols > MAX_X_PIXEL)
    {
      printf("Image too big in x direction \n");
    }
  if ( number_of_rows > MAX_Y_PIXEL )
    {
      printf("Image too big in y direction \n");
    }

  for (jj = 0; jj < number_of_rows; jj += 1)
    {
     for ( ii = 0; ii < number_of_cols; ii += 1)
      {
	//	pixel_array[ii][number_of_rows - jj].greyval = *greydata;
	//   greydata++;
      }
    }

  printf("Read the pgm file, numcols = %d numrows = %d \n", numcols
	 , numrows );

}


void do_prelim()
{
  int ii,jj;

  int x_div_factor;
  int y_div_factor;
  int x_off;
  int y_off;
  int debug;

  debug = 0;

  x_div_factor = 40;
  y_div_factor = 40;

  if (debug) { printf("In prelim \n"); }

  // figure out where the symbol is

   
  dark_threshold = 25;

  box_x_incr = number_of_cols / x_div_factor;
  box_y_incr = number_of_rows / y_div_factor;

  x_off = 0;

  dark_sqindex = 0;

  for(jj = 0; jj < x_div_factor; jj += 1)
    {
      y_off = number_of_rows;

      for(ii = y_div_factor; ii > -1; ii -= 1)
        {
          find_dark_squares( x_off, y_off, box_x_incr, box_y_incr);
          y_off = y_off - box_y_incr;
        }
      x_off = x_off + box_x_incr;
    }

  find_dark_sq_box();

  if (debug) 
     { 
       printf("Found the box around the symbol \n");
       printf("Min box x = %d \n", min_box_x);
       printf("Min box y = %d \n", min_box_y);
       printf("Max box x = %d \n", max_box_x);
       printf("Max box y = %d \n", max_box_y);
     }

  get_box_len();

  //  write_pgm_out(number_of_cols,number_of_rows); 

  init_slopes();   // set up an array of line splopes

  get_scan_points();

  save_line_cnt = 0;

  // get the scan lines, starting with the center 

  fig_flag = FALSE;


  get_scan_lines( scan_cenleft_x, scan_cenleft_y, 1);
  get_scan_lines( scan_upleft_x , scan_upleft_y, 0);
  get_scan_lines( scan_lowleft_x, scan_lowleft_y, 2);


  // no good scans found so try 1/2 and 3/4 points

  if ( fig_flag == FALSE)
    {
    get_scan_lines( scan_upleft_x, ((scan_upleft_y + scan_cenleft_y)/2), 0);
    get_scan_lines( scan_lowleft_x, ((scan_lowleft_y + scan_cenleft_y)/2), 2);
    }
  // get the best qualifying line


  pb_index = 0;

  if ( fig_flag == TRUE)
    {


     do_best(best_pixel_out_xlen, best_pixel_out_ylen);
    }
  else
    {
      printf("No good symbols found using any scan line \n");
      exit(-1);
    }

}  // end do_prelim


// scan program routines start here


// data from the preprocessor
//    %s %s %s = row, col, code
//

void get_preprocessor_data()
{
  int debug;
  int ii;

  int col;
  int row;
  int codeval;
  int tcode_index;
  int rows_from_data;
  int cols_from_data;
  int cval;

  tcode_index = 0;
  erase_count = 0;

  debug = 0;
  rows_from_data = 0;
  cols_from_data = 0;

  codeindex = 0;

  if (debug) { printf("In get frontend processor data \n"); }

  for ( ii = 0; ii < pb_index; ii += 1)
    {
      
      col = pb_array[ii].colval;
      row = pb_array[ii].rowval;

      if ( col > cols_from_data)
        {
	  cols_from_data = col;
        }
      if ( row > rows_from_data)
        {
	  rows_from_data = row;
        }

      codeval = pb_array[ii].codeval;
      cval = (codeval % 929);
      if (debug) { printf("row = %d col = %d code = %d = %d( %d, %d)\n", row, col
			  , codeval, (codeval % 929)
                       ,(cval/ 30), (cval % 30)); }

      if (codeval != 9999)     // 9999 indicates erase or error
	{
         if ( tcode_index < MAX_RAW_CODES )
	  {
           erase_loc[tcode_index] = 0;
          }
        }
      else
        {
         if ( tcode_index < MAX_RAW_CODES )
	  {
           erase_loc[tcode_index] = 1;
           erase_count += 1;
          }
        }
      if ( tcode_index < MAX_RAW_CODES)
       {
        raw_codes[tcode_index].row  = row;
        raw_codes[tcode_index].col  = col;
        raw_codes[tcode_index].code = (codeval % 929);
        if (( row < MAX_ROW) && ( col < MAX_COL) && ( row >  -1) && (col > -1))
         {
           if (codeval != 9999)
	     {
	      rows_cols[row][col+1] = (codeval % 929);
             }
           else
            {
	      rows_cols[row][col+1] = 0;
             }
         }
        else
         {
          if ((row > MAX_ROW) || ( row < 0))
	    {
	      printf("Row value = %d is out of range = %d %d \n", row,
		     0, MAX_ROW);
            }

          if ((col > MAX_COL) || ( col < 0))
	    {
	      printf("Col value = %d is out of range = %d %d \n", col,
		     0, MAX_COL);
            }
	  }
	}
      else
	{
	  printf("Too many raw codes read in \n");
        }
      codeindex += 1;
    }

  

  rows_from_raw = rows_from_data + 1;
  columns_from_raw = cols_from_data + 1;
 

}

 // use a binary search to find the index in the
 // bit_array of the pdf barcode
 //
 int find_index_scan( long inbits)
 {

   int ii;
   int found_index;
   int bit_index;
   int bit_increment;
   int found;
   int less_than_flag;

   int debug;

   bit_index = 1392;

   bit_increment = 696;

   found = FALSE;

   found_index = -1;

   less_than_flag = FALSE;

   debug = FALSE;

   if ( debug ) { printf("Inbits = %x \n", inbits); }

   while (( found == FALSE ) && (bit_increment > 11 ))
     {
       if (debug)
	 {
	     printf(" bit_array val = %x , inbits = %x \n",
				  bit_array[bit_index], inbits); 
	 }
      if (bit_array[bit_index] > inbits)
	{
	 bit_index = bit_index - bit_increment;
	 less_than_flag = FALSE;
	}
      else
	{
	  if (bit_array[bit_index] == inbits)
	    {
	      found = TRUE;
	      found_index = bit_index;
	    }
	  else  // less than 
	    {
	     bit_index = bit_index + bit_increment;
	     less_than_flag = TRUE;
	    }
	}

      bit_increment = bit_increment / 2;

      if (debug)
	{
	 printf("In find loop , bit_increment = %d \n", bit_increment);
	 printf("In find loop , bit_index = %d \n", bit_index);
	}

     }

  if ( found == FALSE)
    {
      ii = 0;
      if ( debug) { printf("no less than flag , bit_index = %d \n",bit_index); }
      while (( found == FALSE) && ( ii < 25) && ( bit_index < 2759))
	{
	   if (bit_array[bit_index + ii] == inbits)
	     {
	      found = TRUE;
	      found_index = bit_index + ii;
	     }
	   ii += 1;        
	}
    }

  if ( found == FALSE)
    {
      ii = 0;
      if (debug) { printf("less than flag , bit_index = %d \n",bit_index); }
      while (( found == FALSE) && ( ii > -25) && ( bit_index > 26))
	{
	  if ( bit_index + ii > 0)
	   {
	    if (bit_array[bit_index + ii] == inbits)
	     {
	      found = TRUE;
	      found_index = bit_index + ii;
	     }
	   }
	   ii -= 1;        
	}
    }

  if (found == FALSE)
    {

      // do an exhaustive search

     ii = 0;

     if (debug) { printf("Doing exhaustive search \n"); }

     while ( ( ii < 2784 ) && (found == FALSE))
       {
	  if ( bit_array[ii] == inbits)
	    {
	     found = TRUE;
	     found_index = ii;
	    }
	   ii += 1;
       }
    }

    if (found == FALSE)
      {
       printf(" Code not found in bit_array = 0x%x \n", inbits);
       return(-1);
      }
    else
      {
       return( found_index );
      }

 }

 // use an exhaustive  search to find the index in the
 // t_array of the pdf barcode
 //
 int find_index_t( int inbits, int *tt_in)
 {

   int ii;
   int found_index;
   int found;
   int kk;  
   int debug;
   int match_found;
  

   found = FALSE;
   debug = 0;

   if ( debug) { printf("Inbits = %x \n", inbits); }
   if ( debug) { printf("tt_in = %d %d %d %d %d %d \n", tt_in[1],
              tt_in[2], tt_in[3], tt_in[4], tt_in[5], tt_in[6]); }
   ii = 0;

   while (( found == FALSE ) && (ii < ( 929 * 3) ))
     {
       match_found = TRUE;
       for ( kk = 1; kk < 7; kk += 1)
	 {
           if (t_arrays[ii][kk] != tt_in[kk])
             {
               match_found = FALSE;
             }
         }
       if ( match_found == TRUE)
         {
	   found = TRUE;
           found_index = ii;
         }
       ii += 1;
     }


    if (found == FALSE)
      {
       printf(" Code not found in t_array = 0x%x \n", inbits);
       return(-1);
      }
    else
      {
       return( found_index );
      }

 }

 int check_correct_row(long in_code, int in_row)
 {
   int this_index;
   int expected_row;
   int bitcode;
   int good_row;

   this_index = find_index_scan( in_code);
   if ( this_index != -1)
     {
      bitcode = bit_codes[this_index]-1;
      expected_row = bitcode /  929;

      good_row = TRUE;
      if ( expected_row != (in_row % 3))
       {
	 printf("Expected row and in_row did not match, exp= %d  in_row= %d \n",
		    expected_row, in_row % 3); 
	good_row = FALSE;
       } 
      return(good_row);
     }
    else
      {
	return(FALSE);
      }

 }

// convert the bit_array values in binary to t_array values
//

void convert_to_t()
{
  int pixel_cnt;
  int ii;
  int mask;
  int color_changes;
  int prev_color;
  int new_color;

  int x_pixel;
  int prev_loc;
  int bar_pwidths[10];
  int bar_widths[10];
  int bits;
  int debug;
  int cluster_num;
 
  double t[10];
  int TT[10];
  int k;
 
  debug = 0;

  for (ii = 0; ii <  929 * 3   ; ii += 1)
    {
       bits= bit_array[ii];
       color_changes = 0;
       pixel_cnt = 0;
       prev_color = (0x10000 & bits ) >> 16;
       mask = 0x10000;
       x_pixel = 0;
       prev_loc = 0;

       for( k = 0; k < 9; k += 1)
	 {
	   bar_widths[k] = 0;
	   bar_pwidths[k] = 0;
	 }

       while( (color_changes < 9) && ( pixel_cnt < 17 ))
	{

	  new_color = (mask & bits) >> ( 16 - pixel_cnt) ;
	  if (new_color != prev_color)
	    {
	      bar_widths[color_changes] = x_pixel - prev_loc;

	      color_changes += 1;
	      prev_loc = x_pixel;
	    }
	  prev_color = new_color;

	  pixel_cnt += 1;
          mask = mask >> 1;
          x_pixel += 1;
	}            

       if (debug) { printf("got bar widths \n"); }

       t[1] = (double) (bar_widths[0] + bar_widths[1]);
       t[2] = (double) (bar_widths[1] + bar_widths[2]);
       t[3] = (double) (bar_widths[2] + bar_widths[3]);
       t[4] = (double) (bar_widths[3] + bar_widths[4]);
       t[5] = (double) (bar_widths[4] + bar_widths[5]);
       t[6] = (double) (bar_widths[5] + bar_widths[6]);
       t[7] = (double) (bar_widths[6] + bar_widths[7]);

       for ( k = 1; k < 8; k += 1)
	{
	  TT[k] = 0;
	 if (( t[k] >= 1.5) && ( t[k] < 2.5))
	  {
	    TT[k] = 2;
	  }
	if (( t[k] >= 2.5) && ( t[k] < 3.5))
	  {
	    TT[k] = 3;
	  }
	if (( t[k] >= 3.5) && ( t[k] < 4.5))
	  {
	    TT[k] = 4;
	  }
	if (( t[k] >= 4.5) && ( t[k] < 5.5))
	  {
	    TT[k] = 5;
	  }
	if (( t[k] >= 5.5) && ( t[k] < 6.5))
	  {
	    TT[k] = 6;
	  }
	if (( t[k] >= 6.5) && ( t[k] < 7.5))
	  {
	    TT[k] = 7;
	  }
	if (( t[k] >= 7.5) && ( t[k] < 8.5))
	  {
	    TT[k] = 8;
	  }
	if (( t[k] >= 8.5) && ( t[k] < 9.5))
	  {
	    TT[k] = 9;
	  }
	}

       cluster_num = ( TT[1] - TT[2] + TT[5] - TT[6] + 9) % 9;

       if ( (cluster_num / 3) != ((bit_codes[ii]-1) / 929))
	 {
	   printf("Conflict between cluster number and index \n");
           printf("Cluster_num = %d ii = %d bitcode[ii] = %d\n",
              cluster_num, ii, bit_codes[ii] );
         }

       for (k = 1; k < 8; k += 1)
	 {   
          t_arrays[ii][k] = TT[k];
         }

       if (debug) {
 
	 printf("bits = %x TT = %d %d %d %d %d %d %d \n",bits,
                  TT[1], TT[2], TT[3], TT[4], TT[5], TT[6], TT[7]);
         
         }
  
    }  // all 929 * 3

}

void init_powers_of_900()
{
strcpy(powers_of_900[0],"000000000000000000000000000000000000000000001");
strcpy(powers_of_900[1],"000000000000000000000000000000000000000000900");
strcpy(powers_of_900[2],"000000000000000000000000000000000000000810000");
strcpy(powers_of_900[3],"000000000000000000000000000000000000729000000");
strcpy(powers_of_900[4],"000000000000000000000000000000000656100000000");
strcpy(powers_of_900[5],"000000000000000000000000000000590490000000000");
strcpy(powers_of_900[6],"000000000000000000000000000531441000000000000");
strcpy(powers_of_900[7],"000000000000000000000000478296900000000000000");
strcpy(powers_of_900[8],"000000000000000000000430467210000000000000000");
strcpy(powers_of_900[9],"000000000000000000387420489000000000000000000");
strcpy(powers_of_900[10],"000000000000000348678440100000000000000000000");
strcpy(powers_of_900[11],"000000000000313810596090000000000000000000000");
strcpy(powers_of_900[12],"000000000282429536481000000000000000000000000");
strcpy(powers_of_900[13],"000000254186582832900000000000000000000000000");
strcpy(powers_of_900[14],"000228767924549610000000000000000000000000000");
strcpy(powers_of_900[15],"205891132094649000000000000000000000000000000");

strcpy(powers_of_900h[0],"000000000000000000000000000000000000000000001");
strcpy(powers_of_900h[1],"000000000000000000000000000000000000000000384");
strcpy(powers_of_900h[2],"0000000000000000000000000000000000000000C5C10");
strcpy(powers_of_900h[3],"00000000000000000000000000000000000002B73A840");
strcpy(powers_of_900h[4],"000000000000000000000000000000002190C2AB18400");
strcpy(powers_of_900h[5],"0000000000000000000000000000007600EC618141000");

strcpy(powers_of_900h_x10[0],"00000000000000000000000000000000000000000000A");
strcpy(powers_of_900h_x10[1],"000000000000000000000000000000000000000002328");
strcpy(powers_of_900h_x10[2],"0000000000000000000000000000000000000007B98A0");
strcpy(powers_of_900h_x10[3],"0000000000000000000000000000000000001B2849280");
strcpy(powers_of_900h_x10[4],"00000000000000000000000000000000005F79A130A00");
strcpy(powers_of_900h_x10[5],"000000000000000000000000000000014FA79AAEF2800");


strcpy(powers_of_900h_x100[0],"000000000000000000000000000000000000000000064");
strcpy(powers_of_900h_x100[1],"000000000000000000000000000000000000000015F90");
strcpy(powers_of_900h_x100[2],"000000000000000000000000000000000000004D3F640");
strcpy(powers_of_900h_x100[3],"0000000000000000000000000000000000010F92DB900");
strcpy(powers_of_900h_x100[4],"0000000000000000000000000000000003BAC04BE6400");
strcpy(powers_of_900h_x100[5],"0000000000000000000000000000000D1C8C0AD579000");


}

int modbase(int x)
{
  return ( x % (GPRIME -1 ));

}

// initialize table of 3**i for syndrome calculation
//

  void powers_init()
  {
  int ii;
  int power_of_3;
  int debug;
  
   debug = 0;
   power_of_3 = 1;
   Index_of[1] = GPRIME-1;

   for (ii = 0; ii < GPRIME-1; ii += 1)
     {
       powers_of_3[ii] = power_of_3;
       Alpha_to[ii] = power_of_3;
       if ( power_of_3 < GPRIME )
         {
          log_of_3[power_of_3] = ii;
          if ( ii != GPRIME - 1)
	    {
             Index_of[power_of_3] = ii;
            }
         }
       else
         {
	   printf("Internal error:  powers of 3 calculation \n");
         }
       
       if (debug) { printf("pow = %d  ii = %d \n", power_of_3, ii); }
       if (debug) { printf("log = %d \n",ii); }
       power_of_3 = (power_of_3 * 3) % GPRIME;
     }
   Index_of[0] = GPRIME-1;
   Alpha_to[GPRIME-1] = 1;
   Index_of[GPRIME] = A0;
  }

// calculate syndromes for the encoded message + correction
// this is to check to see that the error correction codes are correct
//

void syndromes(int message_len, int synd_count )
{

  int i, j;
  int power_val;
  int synd_val;
 
  int debug;
  debug = 0;

  for ( i =0; i < synd_count; i += 1)
    {

      synd_array[i] =0;
    }


  for ( i =0; i < synd_count; i += 1)
    {

      power_val = powers_of_3[i+1];

      if (debug) { printf("Power val = %d \n", power_val); }
      synd_val = 0;
      for(j=0; j < message_len; j += 1)
	{
          synd_val = (data[j] + synd_val) % GPRIME;
          synd_val = (power_val * synd_val) % GPRIME;
          if (debug) { printf("Sval = %d \n", synd_val); }
        }

      synd_array[i] = synd_val;

      if (debug) { printf("i = %d syndrome = %d \n",i, synd_array[i]); }

    }


}

int strtoint( char *instring)

{

int cval;
int val;
char tchar;

 val = 0;
 if (( instring[0] == '0') && (instring[1] == 'x'))
   {
    instring++;
    instring++;
   }

 while(*instring != NULLCHAR)
   {
   
    tchar = *instring;
   
    if (isalpha( tchar))
      {
       cval = tchar - 'a' + 10;
      }
    else
      {
       cval = tchar - '0';
      }
    
    val = (val * 16 ) + cval;
    instring++;
   }
 return(val);
}


void read_raw_bits()
{
int debug;
char numstr[128];
 char tstr[128];
long num;
int bitcode;
int rowval;   // a value from 0-2 indicating that the bit pattern
             // should be in a row such that row%3 = rowval
int incode;

int this_index;
int code_count;
int rows_from_data;
int columns_from_data;

int ii, jj;
int raw_index;
int begin_flag;
 int rows_read;   // number of rows read in
 int col_read;

 infile1=fopen("infiley","r");

 if ( infile1 == NULL)
   {
    printf("Could not open the input file = infiley \n");
    exit(0);
   }

 debug = 0;
 code_count = 0;
 erase_count = 0;

 for ( ii = 0; ii < 1024; ii += 1)
   {
     erase_loc[ii] = 0;
   }

 rows_read = -1;
 col_read = 0;
 while( fscanf(infile1, "%s", numstr) == 1)
   {
    if (debug)
      {
       printf("In read_raw_bits loop = %s \n",numstr);
      }

    begin_flag = FALSE;

     if (( numstr[0] == 'b') || (numstr[0] == 'B'))
       {
         tstr[0] = NULLCHAR;
         for (jj = 1; jj < strlen(numstr); jj += 1)
           {
             tstr[jj-1] = numstr[jj];
           }
         tstr[jj-1] = NULLCHAR;
         strcpy(numstr,tstr);
         begin_flag = TRUE;      // marks beginning of each row
         rows_read += 1;
         col_read = 0;
       }
          
     num = strtoint(numstr);;   // convert from hex?
 
     if (debug) { 
         printf("int value = %d hex = %x \n",num, num); 
         printf(" Numstr = %s num = %x \n", numstr, num);
     }

    // find the binary pattern in the array of all pdf417 patterns

    this_index = find_index_scan( num);

    if ( this_index != -1)
      {
       if (( this_index < 2788) && ( this_index > -1 ))
	 {
	   // find the code that corresponds to this pattern

	   bitcode = bit_codes[this_index]-1;
          if (( bitcode < 0) || ( bitcode > 2787 ))
	    {
               printf(" Bad bitcode = %d \n", bitcode);
            }
          rowval = bitcode / 929;
          if ( rows_read %3 != rowval)
            {
           printf("Error: bitcode is in wrong row : row = %d bitcode = %d \n",
		  rows_read, bitcode );
            }

          incode = bitcode % 929;
          if (debug) 
	    {
          printf("numstr = %s bitcode = %d , rowval = %d, incode = %d \n",
                   numstr,
                   bitcode, rowval, incode );
            }
          if (code_count == 0)
	    {
              rows_from_data = ((( incode % 30) * 3) + 1);
	    }
          raw_codes[code_count].code= incode;
          raw_codes[code_count].row = rows_read;
          raw_codes[code_count].col = col_read;
          code_count += 1;
         }
      }
    else
      {
	// couldn't find the code in the allowed ones
        // count as an erasure
        erase_count += 1;
        erase_loc[code_count] = 1;
        code_count += 1;
      }
    col_read += 1;    
   }

 if (debug)
   {
    printf("In read_raw_bits - rows = %d \n", rows_from_data );
    printf("In read_raw_bits - total codes = %d \n", code_count);
   }
 if ( code_count % rows_from_data != 0 )
  {
   if ( code_count % (rows_from_data + 1) != 0)
     {
       if ( code_count % (rows_from_data + 2) != 0)
         {
          printf("Code count is not a multiple of rows \n");
         }
       else
         {
           rows_from_data += 2;
           columns_from_data = code_count / rows_from_data ;
           if (debug) { printf("Columns from data = %d \n", 
			       columns_from_data ); }
         }
     }
   else
     {
      rows_from_data +=1;
      columns_from_data = code_count / rows_from_data ;
      if (debug) { printf("Columns from data = %d \n", columns_from_data ); }
     }
  }
 else
  {
   columns_from_data = code_count / rows_from_data ;
   if (debug) { printf("Columns from data = %d \n", columns_from_data ); }
  }

 if (debug)
   {
 printf("In read_raw_bits - corrected rows = %d \n", rows_from_data );
 printf("In read_raw_bits -        columns = %d \n", columns_from_data );
   }

 rows_from_raw = rows_from_data;
 columns_from_raw = columns_from_data;

 if (( rows_from_data < MAX_ROW) && ( columns_from_data < MAX_COL))
   {
    raw_index = 0;
    for (ii = 0; ii < rows_from_data; ii += 1)
    {
       for (jj = 1; jj < columns_from_data +1; jj += 1)
        {
          if ( raw_index < MAX_RAW_CODES)
	    {
             rows_cols[ii][jj] = raw_codes[raw_index].code;
             if (debug) { printf("ii = %d jj = %d rows_col[ii][jj] = %d \n",
				 ii,jj, rows_cols[ii][jj]); }
             raw_index += 1;
            }
          else
            {
	      printf("Number of raw codes exceeds max = %d \n",MAX_RAW_CODES);
            }
        }
     }
   }
 else
   {
    printf("Raw data shows too large rows/columns number \n");
   }
 
}

void read_raw()
{

char rowcolstr[120];
char codestr[120];
char numstr[120];
int num;
int ii,jj;
int rowval;
int colval;
char codes[120];
char rows[120];
char cols[120];
int debug;
int code_count;
int actual_row;
int lastcolval;


 infile1=fopen("infilex","r");

 if ( infile1 == NULL)
   {
    printf("Could not open the input file = infilex \n");
    exit(0);
   }

 code_count = 0;

 debug = 0;
 actual_row = -1;

 lastcolval = 9999;
 
 while( fscanf(infile1, "%s %s %s",rowcolstr, codestr, numstr) == 3)
   {
 
    strcpy(rows, rowcolstr);
    ii = 0;
    while ((rowcolstr[ii] != ',') && ( ii < 120) && 
                      (rowcolstr[ii] !=NULLCHAR))
     {
       ii += 1;

     } 
    rows[ii] = NULLCHAR;
    rowval = atoi(rows);

    if (debug) { printf("In read_raw, rows = %s  rowval = %d \n",
			rows, rowval ); }

    ii = ii + 1;
    jj = 0;
    while ((rowcolstr[ii] != NULLCHAR) && ( ii < 120) && 
              ( rowcolstr[ii] != NULLCHAR ))
      {
       cols[jj] = rowcolstr[ii];
       ii += 1;
       jj += 1;
      }
    cols[jj] = NULLCHAR;

    colval = atoi(cols);


    if (colval < lastcolval)
      {
        actual_row += 1;
      }

    lastcolval = colval;

    if (debug) { printf("In read_raw, cols = %s  colval = %d \n",
			cols, colval ); }

    strcpy(codes, codestr);
    ii = 0;

    while (( codes[ii] != '-') && ( ii < 120) && (codes[ii] != NULLCHAR))
     
      {
       ii += 1;
      }
    codes[ii] = NULLCHAR;


    if (debug) { printf("In read_raw, codes = %s  numstr  = %s a_row = %d\n",
			codes, numstr , actual_row); }
  
    num = atoi(numstr);
    
    if (( actual_row < MAX_ROW) && ( colval < MAX_COL))
      {
       rows_cols[actual_row][colval] = num;
      }

     code_count += 1;

   }

 printf(" number of codes = %d \n", code_count );

 fclose( infile1 );

}  /* read raw file */

int get_next_code( )
{
int outval;
int debug;


 debug = 0;


 outval = code_array[codeindex];

 if (debug)
   {
     printf("In get_next_code - code = %d codeindex = %d\n", outval,codeindex);
     printf("  code, l,r = %d %d \n", outval/30, outval%30);
     printf(" code_array_size = %d number_check_words = %d \n",
	    code_array_size, number_check_words );
   } 
 
 if ((codeindex > 0 ) && ( codeindex == code_array_size -1 - number_check_words))
   {
     if ( debug) { printf("Codeindex points to check words \n"); }
     codeindex += 1;
     return(999);
   }
 
 if (codeindex < MAX_CODE )
   {
    codeindex += 1;
   }
 else
   {
    printf(" Too many code words \n");
   }

 return( outval);

}

void out_byte ( int inval)
{
 
fprintf(outfile1,"%x", inval);

} /* do out_byte */

void do_text_compress()
{
int byte_index;

int ii;
 int kk;

int text_compress_end;
int byte;
char char1;
char char_array[1024];
int char_index;
int mode_change;
 int save_mode;
int debug;
 int j;
 int in_loop;
int sub_mode;
 int dquotchar;
 char outstr[20];


   debug = 0;

   byte_index = 0;

   char_index = 0;

   text_compress_end = FALSE;

   if (debug) { printf("In text compress: code = %d \n", code); }

   in_loop = FALSE;                
   while( text_compress_end == FALSE)
     {

       if ((code == 900) && ( in_loop == FALSE))
         {
           code = get_next_code(); 
      
         }

       in_loop = TRUE;
       if (code < 900)
	 {
           byte_array[byte_index]  = code / 30;
           byte_array[byte_index+1] = code % 30;
           byte_index += 2;
           if ( byte_index == MAX_BYTE_INDEX)
             {
              printf("Error:  Too many text compress chars \n");

             }
         }
       else
	 {
           if ( code == 900 )
	     {
               text_compress_end = TRUE;
               text_compress_mode = TRUE;
             }
           if ( code == 901 )
	     {
               text_compress_end = TRUE;
               byte_compress_mode = TRUE;
             }

           if ( code == 924 )
	     {
               text_compress_end = TRUE;
               byte_compress_mode = TRUE;
             }
           if ( code == 902 )
	     {
               text_compress_end = TRUE;
               numeric_compress_mode = TRUE;
             }
           if ( code == 999 )
	     {
               text_compress_end = TRUE;
               end_found = TRUE;
             }
            if ( code == 913 )
 	     {
                code = get_next_code();

                sprintf(outstr,"%x", code);
                
		byte_array[byte_index] = 913;
                 
                strcpy( byte_913_str[byte_index].str913, outstr);

                byte_index += 1;
                byte_compress_mode = TRUE;
              }
	 }

      if ( text_compress_end == FALSE)
	{
         code = get_next_code();
        }
      if ( debug)
	{
         printf("In text_compress loop - code = %d (%d,%d) \n", code
                       , code/30, code%30 );
        }


     }


   if (debug) { printf(" byte_index = %d \n", byte_index); }

   if (debug) { for (ii = 0 ; ii < byte_index; ii += 1)
                    {
		      printf("byte_array[ii] = %d ii = %d \n",
                              byte_array[ii],ii); 
                    }
               }

 sub_mode = ALPHA;
 save_mode = ALPHA;

 ii = 0;

 if (debug ) { printf(" Byte index = %d \n", byte_index); }

 while ( ii < byte_index)
   {
    byte = byte_array[ii];

    if (debug) { printf("In byte loop - byte = %d \n", byte); }
      
    char1 = NULLCHAR;

    switch( sub_mode )
          {
	  case ALPHA :
	    if (debug) 
              { printf(" In alpha mode \n"); }
            mode_change = FALSE;
      
            if ( byte < 26 )
              {
                char1 = 'A' + byte;
              }
            else
              {
                 if ( byte == 26 )
		   {
                      char1 = ' ';
                   }
                 if ( byte == 27 )
		   {
                      sub_mode = LOWER;
                      mode_change = TRUE;
                    }
                 if ( byte == 28 )
		   {
                      sub_mode = MIXED;
                      mode_change = TRUE;   
                   }
                 if ( byte == 29 )
		   {
		     save_mode = sub_mode;
                      sub_mode = PUNCT_SH;
                      mode_change = FALSE;
                   }
		 if ( byte == 913)
                   {
                     strcpy( outstr, byte_913_str[ii].str913);
                      
		     for ( kk = 0; kk < strlen( outstr); kk += 1)
                      {
			char_array[char_index] = outstr[kk];
                        char_index += 1;
                      }
                   }             
	      }
            break;

      case LOWER :
	    
            mode_change = FALSE;
      
	    if (debug) 
              { printf(" In Lower mode \n"); }
            if ( byte < 26 )
              {
                char1 = 'a' + byte;
              }
            else
              {
                 if ( byte == 26 )
		   {
                      char1 = ' ';
                   }
                 if ( byte == 27 )
		   {
                      sub_mode = ALPHA;
                      mode_change = TRUE;
                    }
                 if ( byte == 28 )
		   {
                      sub_mode = MIXED;  
                      mode_change = TRUE;                 
                   }
                 if ( byte == 29 )
		   {
		     save_mode = sub_mode;
                      sub_mode = PUNCT_SH;
                      mode_change = FALSE;
                   }

		 if ( byte == 913)
                   {
                     strcpy( outstr, byte_913_str[ii].str913);
                      
		     for ( kk = 0; kk < strlen( outstr); kk += 1)
                      {
			char_array[char_index] = outstr[kk];
                        char_index += 1;
                      }
                   }
	      }
	  
            break;

     case MIXED :
         
            mode_change = FALSE;
      
	    if (debug) 
              { printf(" In mixed mode \n"); }

            if ( byte < 25 )
              {
                char1 = mixed_chars[byte];
              }
            else
              {
                 if ( byte == 25 )
		   {
                      sub_mode = PUNCT;
                      mode_change = TRUE;
                   }
                 if ( byte == 26 )
		   {
                     char1 = ' ';
                    }
                 if ( byte == 27 )
		   {
                      sub_mode = LOWER;
                      mode_change = TRUE;
                   }
                 if ( byte == 28 )
                   {
                     sub_mode = ALPHA;
                     mode_change = TRUE;
                   }
                 if ( byte == 29 )
		   {
		     save_mode = sub_mode;
                       sub_mode = PUNCT_SH;
                       mode_change = FALSE;
                    }

		 if ( byte == 913)
                   {
                     strcpy( outstr, byte_913_str[ii].str913);
                      
		     for ( kk = 0; kk < strlen( outstr); kk += 1)
                      {
			char_array[char_index] = outstr[kk];
                        char_index += 1;
                      }
                   }
	      }
            break;

     case PUNCT :
         
            mode_change = FALSE;

	    if (debug) 
              { printf(" In punct mode \n"); }
      
            if ( byte < 29 )
              {
                char1 = punct_chars[byte];
              }
            else
              {
                 if ( byte == 29 )
		   {
                      sub_mode = ALPHA;
                      mode_change = TRUE;
                   }

		 if ( byte == 913)
                   {
                     strcpy( outstr, byte_913_str[ii].str913);
                      
		     for ( kk = 0; kk < strlen( outstr); kk += 1)
                      {
			char_array[char_index] = outstr[kk];
                        char_index += 1;
                      }
                   }
	      }
            break;

     case PUNCT_SH :
         
            mode_change = FALSE;
            sub_mode = save_mode;
	    if (debug) 
              { printf(" In punct_shift mode \n"); }
      
            if ( byte < 29 )
              {
                char1 = punct_chars[byte];
              }
            else
              {
                 if ( byte == 29 )
		   {
                      sub_mode = ALPHA;
                      mode_change = TRUE;
                   }
	      }
            break;
	  }

       if ( char1 != NULLCHAR)
	 {
	   if ( char_index < 1024 )
	     {
              char_array[char_index] = char1;
              char_index += 1;
             }
           else
	     {
	       printf("Error: more than 1024 chars in text compress \n");
             }
          
          char_array[char_index+1] = NULLCHAR;
          if (debug) { printf(" chars = %c \n", char1); }

         }
          
       ii += 1;

       if (debug) { printf("ii = %d \n",ii ); }

     }  /* end while */

 dquotchar = 34;

 if ( char_index > 0 )
   {
    fprintf(outfile1,"\nTC %c", dquotchar);
    for(j = 0; j < char_index ; j += 1)
      {
       fprintf(outfile1,"%c",char_array[j]);
      }
    fprintf(outfile1,"%c\n",dquotchar);
   }

 if (debug) { printf("Leaving text compress \n"); }

}  /* text compress  */

/* just drop the first char, get the rest of string */

void strshift( char *instr1, char *instr2)

{
int ii;
char outstr[256];
char tempstr[256];

 strcpy( tempstr, instr2);

 for ( ii = 0; ii < (strlen(tempstr)) -1; ii += 1)
   {
    outstr[ii] = tempstr[ii+1];
   }

 outstr[ii] = NULLCHAR;

 *instr1 = outstr[0];

}

void do_rows_columns( int rows, int cols )
{
int ii;
int kk;
 int debug;
int index;
int xindex_l;
int xindex_r;
int vindex_l;
int vindex_r;
int yindex_l;
int yindex_r;
int zindex_l;
int zindex_r;
int error_correct_lev;
int num_cols_r;


 debug = 0;
 for ( ii = 0; ii < rows; ii += 1)
   {
    if ( (ii % 3) == 0)
      {
         xindex_l = (rows_cols[ii][1] / 30);
         if ( xindex_l != ii/3 )
           {
             printf(" Bad left indicator in row = %d \n", ii + 1);
             printf(" X value = %d expected value = %d \n", 
                            xindex_l, ii/3);
	   }

         yindex_l = (rows_cols[ii][1] % 30) ;

         
         xindex_r = (rows_cols[ii][cols] / 30);

         vindex_r = (rows_cols[ii][cols] % 30);

         if ( vindex_r != (cols-3) ) // cols = total cols = data reg cols + 2
	   {
            printf(" Bad right indicator in row = %d \n", ii + 1);
            printf(" V value = %d expected value = %d \n", vindex_r, cols-3 );

	   }
         num_cols_r = vindex_r + 1;
         
         if ( yindex_l != (( rows - 1) / 3))
	   {
            printf("Error in left row indicator in row = %d \n",ii + 1);
            printf(" Expected %d and got %d \n", (rows -1 )/3,
                  yindex_l);
           }
         if ( xindex_r != xindex_l)
	   {
            printf(" Mismatch between left and right indicator \n");
            printf(" In row  = %d \n", ii + 1);
            printf(" Left x value = %d right x value = %d \n", xindex_l,
                       xindex_r);
	   }


      }
    if  ( (ii % 3) == 1)
      {
         xindex_l = (rows_cols[ii][1] / 30);
         zindex_l = (rows_cols[ii][1] % 30);

         error_correct_lev = (zindex_l - ((rows -1) % 3)) / 3;

         xindex_r = (rows_cols[ii][cols] / 30);
         yindex_r = (rows_cols[ii][cols] % 30);
 
         if ( yindex_r != yindex_l)
           {
            printf(" Mismatch between left and right indicator \n");
            printf(" In row  = %d \n", ii + 1);
            printf(" Left y value = %d right y value = %d \n", yindex_l,
                       yindex_r);
           }

         if ( xindex_r != xindex_l)
	   {
            printf(" Mismatch between left and right indicator \n");
            printf(" In row  = %d \n", ii + 1);
            printf(" Left x value = %d right x value = %d \n", xindex_l,
                       xindex_r);
	   }

      }
    if ( (ii % 3) == 2)
      {
         xindex_l = (rows_cols[ii][1] / 30);
         vindex_l = (rows_cols[ii][1] % 30);

         xindex_r = (rows_cols[ii][cols] / 30);
         zindex_r = (rows_cols[ii][cols] % 30);
 
         if ( zindex_r != zindex_l)
           {
            printf(" Mismatch between left and right indicator \n");
            printf(" In row  = %d \n", ii + 1);
            printf(" Left z value = %d right z value = %d \n", zindex_l,
                       zindex_r);
           }


         if ( xindex_r != xindex_l)
	   {
            printf(" Mismatch between left and right indicator \n");
            printf(" In row  = %d \n", ii + 1);
            printf(" Left x value = %d right x value = %d \n", xindex_l,
                       xindex_r);
	   }

      }
      
   }

 index = 0;
 for ( ii = 0; ii < rows ; ii += 1 )
   {
    for ( kk = 1; kk < cols + 1; kk += 1)
      {
        if (( kk == 1) || (kk == cols))
          {
          }
        else
	  {
            if ( index < MAX_CODE )
	      {
               code_array[index] = rows_cols[ii][kk];
               if (debug) {
                    printf(" index = %d code_array[index] = %d \n",index,
			   code_array[index]); }
               index += 1;
              }
            else
              {
                printf("Size of row/column array too large \n");
              }
	  }
       }
   }

 number_check_words = 2;

 for ( kk = 0; kk < error_correct_lev ; kk += 1 )
   {
     number_check_words = number_check_words * 2;
   }

 if (debug)
   {
    printf("number of check words = %d \n", number_check_words);
    printf("index = %d \n",index);
   }

 code_array_size = index;

 if (verbose_out)
   {
    printf("code array size = %d \n", code_array_size );
   }

 for ( kk = 0; kk < index; kk += 1)
   {
     data[kk] = code_array[kk];
   }

 syndromes( index, ( number_check_words/2) );

 // fix up the data with Reed Solomon Error correction

 eras_dec_rs(data, erase_loc, erase_count, index,
	     (number_check_words/ 2));


 for ( kk = 0; kk < index; kk += 1)
   {
     if ( data[kk] != code_array[kk] )
       {
	 if (debug)  { printf("Data at %d corrected , was = %d now %d \n", kk , 
			      code_array[kk], data[kk]); }

         code_array[kk] = data[kk];
       }
   }
 
}  /* do row columns */


void init_char_arrays()
{

int debug;
debug = 0;

if (debug )
  {
   printf("In init_char_arrays \n");
  }
punct_chars[0] = ';';
punct_chars[1] = '<';
punct_chars[2] = '>';
punct_chars[3] = '@';
punct_chars[4] = '[';
punct_chars[5] = 92;
punct_chars[6] = '}';
punct_chars[7] = '_';
punct_chars[8] = 96;
punct_chars[9] = '~';
punct_chars[10] = '!';
punct_chars[11] = 13;
punct_chars[12] = 9;
punct_chars[13] = ',';
punct_chars[14] = ':';
punct_chars[15] =  10;
punct_chars[16] = '-';
punct_chars[17] = '.';
punct_chars[18] = '$';
punct_chars[19] = '/';
punct_chars[20] = 34;
punct_chars[21] = '|';
punct_chars[22] = '*';
punct_chars[23] = '(';
punct_chars[24] = ')';
punct_chars[25] = '?';
punct_chars[26] = '{';
punct_chars[27] = '}';
punct_chars[28] = 39;

mixed_chars[0] = '0';
mixed_chars[1] = '1';
mixed_chars[2] = '2';
mixed_chars[3] = '3';
mixed_chars[4] = '4';
mixed_chars[5] = '5';
mixed_chars[6] = '6';
mixed_chars[7] = '7';
mixed_chars[8] = '8';
mixed_chars[9] = '9';
mixed_chars[10] = '&'; 
mixed_chars[11] = punct_chars[11];
mixed_chars[12] = punct_chars[12];
mixed_chars[13] = punct_chars[13];
mixed_chars[14] = punct_chars[14];
mixed_chars[15] = '#';
mixed_chars[16] = punct_chars[16];
mixed_chars[17] = punct_chars[17];
mixed_chars[18] = punct_chars[18];
mixed_chars[19] = punct_chars[19];
mixed_chars[20] = '+';
mixed_chars[21] = '%';
mixed_chars[22] = punct_chars[22];
mixed_chars[23] = '=';
mixed_chars[24] = '^';

if (debug)
  {
   printf("Leaving init_char_arrays \n");
  }


}

/* add two 45 char strings */

void add_char_45( char *out1, char *in1, char *in2)
{

int carval;
int ii;
int intval1, intval2;
int sumval;

char instring1[256];
char instring2[256];
int debug;

 carval = 0;

 debug = 0;

 if (debug) { printf("in1 = %s in2 = %s \n", in1, in2); }

 for ( ii = 14 ; ii >  -1 ; ii -= 1)
   {
     instring1[0] = in1[ ii * 3];
     instring1[1] = in1[ ( ii * 3 ) + 1];
     instring1[2] = in1[ ( ii * 3 ) + 2];
     instring1[3] = NULLCHAR;

     instring2[0] = in2[ ii * 3];
     instring2[1] = in2[ ( ii * 3 ) + 1];
     instring2[2] = in2[ ( ii * 3 ) + 2];
     instring2[3] = NULLCHAR;

     intval1 = atoi(instring1);
     intval2 = atoi(instring2);

     if (debug) { printf("Intval1 = %x Intval2 = %x \n", intval1, intval2); }

     sumval = ( intval1 + intval2 + carval) % 1000;
     
     carval = (intval1 + intval2 + carval) / 1000;
     
     out1[ (ii * 3) + 2] = (sumval % 10) + '0';
     out1[ ( ii * 3) + 1] = (( sumval / 10) % 10) + '0';
     out1[ ( ii * 3) ] = ( sumval / 100) + '0';
   }

 if (debug)
   {
    printf("In add char45:  \n");
    printf("In1 = %s \n", in1);
    printf("In2 = %s \n", in2);
    printf("out = %s \n", out1);
   }
}

/* add two 45 char hex strings */

void add_char_45hex( char *out1, char *in1, char *in2)
{

int ii;
int carval;
int intval1;
int intval2;
int sumval;
int sv1, sv2, sv3;
char instring1[32];
char instring2[32];

int debug;

 carval = 0;

 debug = 0;
 if (debug) { printf("in1 = %s in2 = %s \n", in1, in2); }

 for ( ii = 14 ; ii >  -1 ; ii -= 1)
   {
     instring1[0] = in1[ ii * 3];
     instring1[1] = in1[ ( ii * 3 ) + 1];
     instring1[2] = in1[ ( ii * 3 ) + 2];
     instring1[3] = NULLCHAR;

     instring2[0] = in2[ ii * 3];
     instring2[1] = in2[ ( ii * 3 ) + 1];
     instring2[2] = in2[ ( ii * 3 ) + 2];
     instring2[3] = NULLCHAR;

     intval1 = (int ) strtol(instring1, NULL, 16);
     intval2 = (int ) strtol(instring2, NULL, 16);

     if (debug) { printf("Intval1 = %x Intval2 = %x \n", intval1, intval2); }
    
     sumval = ( intval1 + intval2 + carval) % 4096;
     
     carval = (intval1 + intval2) / 4096;
     
     sv1 = sumval % 16;
     if ( sv1 < 10)
       {

        out1[ (ii * 3) + 2] = sv1 + '0';
       }
     else
       {
        out1[ (ii * 3) + 2] = sv1 - 10 + 'A';
       }
     sv2 = (( sumval / 16 ) % 16 );
     if ( sv2 < 10)
       {
        out1[ ( ii * 3) + 1] = sv2 + '0';
       }
     else
       {
        out1[ (ii * 3) + 1] = sv2 - 10 + 'A';
       }
     sv3 = sumval / 256;
     if ( sv3 < 10)
       {
         out1[ ( ii * 3) ] = sv3 + '0';
       }
     else
       {
         out1[ ( ii * 3) ] = sv3 - 10 +  'A';
       }

   }
 if (debug)
   {
    printf("In add char45:  \n");
    printf("In1 = %s \n", in1);
    printf("In2 = %s \n", in2);
    printf("out = %s \n", out1);
   }
}


/*  take base 900 value in codes_bc and convert to the  base 16 */
/* string in result */

void do_900_to_16()

{
int kk;
int hundreds;
int tens;
int ones;
char resultc[80];
int jj;
int debug;

char tstr[80];
char addend[64];


  debug = 0;

  if (debug) { printf("In do_900_to_16 \n"); }

  for ( kk = 0; kk < 60; kk += 1)
    {
     result[kk] = '0';
    }

  resultc[0] = NULLCHAR; 
  
  result[45] = NULLCHAR;

  for ( jj = 4; jj > -1; jj -= 1)
    {
      if ((codes_bc[jj] > 900 ) || (codes_bc < 0))
	{
          printf("Illegal code value for 900 to hex conversion \n");
          printf(" Illegal value = %d at code position = %d \n", 
		 codes_bc[jj], jj );
          return;
  
        }
    }
  for ( jj = 4; jj > -1; jj -= 1)
    {
       hundreds = codes_bc[jj] / 100;
       tens  = (codes_bc[jj] / 10) % 10;
       ones = codes_bc[jj] % 10;

       if ( debug) 
          { 
            printf("Hundreds = %d tens = %d ones = %d \n", 
                                     hundreds, tens, ones );
          }
       strncpy( addend, powers_of_900h[4-jj],45);
       addend[45]= NULLCHAR;
      
       if (debug ) { printf("Addend = %s \n", addend); }
  
       for ( kk = 0; kk < ones ; kk += 1)
	 {
           strcpy(resultc, result);

           add_char_45hex( result, addend, resultc);
         }
       
       strncpy( tstr, powers_of_900h_x10[ 4 -jj],45);
       tstr[45]= NULLCHAR;

       if ( debug ) { printf("Tstr = %s \n", tstr); }

       for ( kk = 0; kk < tens ; kk += 1)
	 {
           strcpy(resultc, result);
           add_char_45hex( result, tstr, resultc);
         }

       strncpy( tstr, powers_of_900h_x100[ 4-jj],45);
       tstr[45]= NULLCHAR;

       if ( debug ) { printf("Tstr = %s \n", tstr); }

       for ( kk = 0; kk < hundreds ; kk += 1)
	 {
           strcpy(resultc, result);
           add_char_45hex( result, tstr , resultc);
         }
       
     
    }

  if (debug)
    {
     printf("Leaving do_900_to_16 - result = %s \n", result);
    }

} /* do 900 to 16 */

void print_result()
{
int ii;

 ii = 0;

 fprintf(outfile1,"\nBC ");
 while (( result[ii] == '0') && ( ii < 50))
  {
   ii += 1;
  }

 while  ( ii < 50)
  {
    if ( result[ii] != 13)
      {
       fprintf(outfile1,"%c",result[ii]);
      }
    else
      {
	fprintf(outfile1,"\n");
      }

   ii += 1;
  }

} /* print_result */

/* take base 900 value in codes_nc and convert to the  base 10 */
/* string in result */

void do_900_to_10()

{
int kk;
 int mm;
int hundreds;
int tens;
int ones;
char resultc[128];
int jj;
char shstr[128];
int debug;
char tstr[128];
 char addend[64];

 debug = 0;

  for ( kk = 0; kk < 45; kk += 1)
    {
     result[kk] = '0';
    }

  result[kk] = NULLCHAR;

  resultc[0] = NULLCHAR; 

  for ( jj = 14; jj > -1; jj -= 1)
    {
      if ((codes_nc[jj] > 900) || ( codes_nc[jj] < 0))  
       {
        printf("Illegal value in numerical compress codes = %d at loc = %d \n",
		 codes_nc[jj] , jj );
        return;

       }
    }
  for ( jj = 14; jj > -1; jj -= 1)
    {
      if (debug) { printf("jj = %d \n", jj); }

       hundreds = codes_nc[jj] / 100;
       tens  = (codes_nc[jj] / 10) % 10;
       ones = codes_nc[jj] % 10;

       strncpy( addend, powers_of_900[14-jj], 45);

       addend[45] = NULLCHAR;

       if (debug) { printf("addend = %s \n", addend); }
       
       for ( kk = 0; kk < ones ; kk += 1)
	 {
           strncpy(resultc, result,45);
           add_char_45( result, addend, resultc);
           
         }
       
       strncpy( tstr, powers_of_900[ 14 -jj],45);
       tstr[45] = NULLCHAR;

       strcat ( tstr, "0");
       
       //       strshift( shstr, tstr);       
       for(mm = 0; mm < strlen(tstr); mm += 1)
         {
           shstr[mm] = tstr[mm+1];
         }
       shstr[mm] = NULLCHAR;
                        
       if ( debug ) 
          { 
            printf("Tstr = %s \n", tstr);
            printf("shstr = %s \n", shstr);
           }


       for ( kk = 0; kk < tens ; kk += 1)
	 {
           strncpy(resultc, result,45);

           add_char_45( result, shstr, resultc);
           
         }
     
       strcat ( tstr, "0");
     
       //       strshift ( shstr, tstr);
       for(mm = 0; mm < strlen(tstr)-1; mm += 1)
         {
           shstr[mm] = tstr[mm+2];
         }
       shstr[mm] = NULLCHAR;

       if ( debug ) { printf("shstr = %s \n", shstr); }

       for ( kk = 0; kk < hundreds ; kk += 1)
	 {
           strncpy(resultc, result, 45);
           add_char_45( result, shstr , resultc);
         }
    }

 fprintf(outfile1,"\nNC %s \n", result);

  if ( debug) 
    {
     printf("In do_900_to_10 - result = %s \n",result);
    }

} /* do 900 to 10 */

void do_byte_compress( int style)
{

int byte_compress_codes[ 1024];
int ii;
int end_byte_compress;
int count;
 int debug;

 debug = 0;
 if (debug) { printf("In do_byte_compress - style = %d \n",style); }

 end_byte_compress = FALSE;

 if ( style == 901)  /* codewords not divisible by 6 */
   {
    count = 0;
 
    while(( codeindex < code_array_size) && ( end_byte_compress == FALSE))
     {

       code = get_next_code();
       if ( code < 900 )
        {
          byte_compress_codes[count] = code;
          count += 1;
        }
       else
        {
         if ( ( code == 900) || ( code == 903))
	   {
             end_byte_compress = TRUE;
           }
        }
  
      if (( count % 5 == 0) && ( count > 0))
        {
          codes_bc[0] = byte_compress_codes[count-4];
          codes_bc[1] = byte_compress_codes[count-3];
          codes_bc[2] = byte_compress_codes[count-2];
          codes_bc[3] = byte_compress_codes[count-1];
          codes_bc[4] = byte_compress_codes[count-0];
          do_900_to_16();
          print_result();
          
        }         
     }
    /* put out the remainder of the codes directly */

    for ( ii = (( count / 5) * 5); ii < count; ii += 1)
      {
  
       out_byte( byte_compress_codes[ii] );

      }
   }   /* 901 style */


 if ( style == 913)  /* single codeword follows*/
   {
 
    while(( codeindex < code_array_size) && ( end_byte_compress == FALSE))
     {

       count = 0;
       code = get_next_code();
       end_byte_compress = TRUE;

       if ( code < 900 )
        {
          byte_compress_codes[count] = code;
          count += 1;
        }
       else
        {
         if ( ( code == 900) || ( code == 903))
	   {
             end_byte_compress = TRUE;
           }
        }
      }
     out_byte ( byte_compress_codes[0] );

   }  /* Modes shift to byte_compress mode */


 if ( style == 924)  /* codewords divisible by 6 */
   {
 
    count = 0;
    while(( codeindex < code_array_size) && ( end_byte_compress == FALSE))
     {

       code = get_next_code();
       if ( code < 900 )
        {
          byte_compress_codes[count] = code;
          count += 1;
        }
       else
        {
         if  ( code > 899 )
	   {
             end_byte_compress = TRUE;
             if (( code != 900) && ( count != 903))
	       {
                 printf("Unexpected code in byte compress = %d \n", code);
                 code = get_next_code();
               }
           }
        }
  
      if (( count % 5 == 0) && ( count > 0))
        {
          codes_bc[0] = byte_compress_codes[count-4];
          codes_bc[1] = byte_compress_codes[count-3];
          codes_bc[2] = byte_compress_codes[count-2];
          codes_bc[3] = byte_compress_codes[count-1];
          codes_bc[4] = byte_compress_codes[count-0];
          do_900_to_16();
          print_result();
          
        }         
     }
   }

}

void test_900tohex()
{

          codes_bc[0] = 20;
          codes_bc[1] = 0;
          codes_bc[2] = 0;
          codes_bc[3] = 0;
          codes_bc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();


          codes_bc[0] = 0;
          codes_bc[1] = 20;
          codes_bc[2] = 0;
          codes_bc[3] = 0;
          codes_bc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 0;
          codes_bc[1] = 0;
          codes_bc[2] = 20;
          codes_bc[3] = 0;
          codes_bc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 0;
          codes_bc[1] = 0;
          codes_bc[2] = 0;
          codes_bc[3] = 20;
          codes_bc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 0;
          codes_bc[1] = 0;
          codes_bc[2] = 0;
          codes_bc[3] = 0;
          codes_bc[4] = 20;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 899;
          codes_bc[1] = 0;
          codes_bc[2] = 0;
          codes_bc[3] = 0;
          codes_bc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();


          codes_bc[0] = 0;
          codes_bc[1] = 899;
          codes_bc[2] = 0;
          codes_bc[3] = 0;
          codes_bc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 0;
          codes_bc[1] = 0;
          codes_bc[2] = 899;
          codes_bc[3] = 0;
          codes_bc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 0;
          codes_bc[1] = 0;
          codes_bc[2] = 0;
          codes_bc[3] = 899;
          codes_bc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 0;
          codes_bc[1] = 0;
          codes_bc[2] = 0;
          codes_bc[3] = 0;
          codes_bc[4] = 899;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 0;
          codes_bc[1] = 0;
          codes_bc[2] = 899;
          codes_bc[3] = 899;
          codes_bc[4] = 899;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 0;
          codes_bc[1] = 0;
          codes_bc[2] = 0;
          codes_bc[3] = 899;
          codes_bc[4] = 899;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 0;
          codes_bc[1] = 899;
          codes_bc[2] = 899;
          codes_bc[3] = 899;
          codes_bc[4] = 899;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 901;
          codes_bc[1] = 0;
          codes_bc[2] = 0;
          codes_bc[3] = 0;
          codes_bc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();


          codes_bc[0] = 0;
          codes_bc[1] = 901;
          codes_bc[2] = 0;
          codes_bc[3] = 0;
          codes_bc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 0;
          codes_bc[1] = 0;
          codes_bc[2] = 901;
          codes_bc[3] = 0;
          codes_bc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 0;
          codes_bc[1] = 0;
          codes_bc[2] = 0;
          codes_bc[3] = 901;
          codes_bc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();

          codes_bc[0] = 0;
          codes_bc[1] = 0;
          codes_bc[2] = 0;
          codes_bc[3] = 0;
          codes_bc[4] = 901;

          printf("codes = %d %d %d %d %d \n", codes_bc[0], codes_bc[1],
                         codes_bc[2], codes_bc[3], codes_bc[4]);

          do_900_to_16();
 
}

void test_900to10()
{
  int jj;
  for (jj = 14; jj > -1; jj -=1)
    {
      codes_nc[jj] = 0;
    }

           codes_nc[14] = 879;
           codes_nc[13] = 321;
           codes_nc[12] = 124;
           printf("codes = %d %d %d \n",
		   codes_nc[12], codes_nc[13], codes_nc[14]);
          do_900_to_10();

	  codes_nc[14] = 0;
          codes_nc[13] = 0;
          codes_nc[12] = 0;

          codes_nc[0] = 20;
          codes_nc[1] = 0;
          codes_nc[2] = 0;
          codes_nc[3] = 0;
          codes_nc[4] = 0;
           printf("codes = %d %d %d %d %d \n", codes_nc[0], codes_nc[1],
                         codes_nc[2], codes_nc[3], codes_nc[4]);
          do_900_to_10();

          codes_nc[0] = 0;
          codes_nc[1] = 20;
          codes_nc[2] = 0;
          codes_nc[3] = 0;
          codes_nc[4] = 0;
 
          printf("codes = %d %d %d %d %d \n", codes_nc[0], codes_nc[1],
                         codes_nc[2], codes_nc[3], codes_nc[4]);

          do_900_to_10();

          codes_nc[0] = 0;
          codes_nc[1] = 0;
          codes_nc[2] = 20;
          codes_nc[3] = 0;
          codes_nc[4] = 0;

          printf("codes = %d %d %d %d %d \n", codes_nc[0], codes_nc[1],
                         codes_nc[2], codes_nc[3], codes_nc[4]);
          do_900_to_10();


          codes_nc[0] = 0;
          codes_nc[1] = 0;
          codes_nc[2] = 0;
          codes_nc[3] = 20;
          codes_nc[4] = 0;



          printf("codes = %d %d %d %d %d \n", codes_nc[0], codes_nc[1],
                         codes_nc[2], codes_nc[3], codes_nc[4]);
          do_900_to_10();

          codes_nc[0] = 0;
          codes_nc[1] = 0;
          codes_nc[2] = 0;
          codes_nc[3] = 0;
          codes_nc[4] = 20;

          printf("codes = %d %d %d %d %d \n", codes_nc[0], codes_nc[1],
                         codes_nc[2], codes_nc[3], codes_nc[4]);
          do_900_to_10();


          codes_nc[0] = 0;
          codes_nc[1] = 0;
          codes_nc[2] = 0;
          codes_nc[3] = 20;
          codes_nc[4] = 20;

          printf("codes = %d %d %d %d %d \n", codes_nc[0], codes_nc[1],
                         codes_nc[2], codes_nc[3], codes_nc[4]);
          do_900_to_10();
 
}

void do_nc_compress()
{

int nc_compress_codes[ 1024];

int count;
int end_nc_compress;

 end_nc_compress = FALSE;


    count = 0; 
    while(( codeindex < code_array_size) && ( end_nc_compress == FALSE))
     {

       code = get_next_code();

       if ( code < 900 )
        {
          nc_compress_codes[count] = code;
          count += 1;
        }
       else
        {
          if ( ( code == 900 ) || ( code == 924) || (code == 901))
	    {
              end_nc_compress = TRUE;
            }
      
        }
  
      if ( (count % 15) == 0)
        {
          codes_nc[0] = nc_compress_codes[count-14];
          codes_nc[1] = nc_compress_codes[count-13];
          codes_nc[2] = nc_compress_codes[count-12];
          codes_nc[3] = nc_compress_codes[count-11];
          codes_nc[4] = nc_compress_codes[count-10];

          codes_nc[5] = nc_compress_codes[count-9];
          codes_nc[6] = nc_compress_codes[count-8];
          codes_nc[7] = nc_compress_codes[count-7];
          codes_nc[8] = nc_compress_codes[count-6];
          codes_nc[9] = nc_compress_codes[count-5];

          codes_nc[10] = nc_compress_codes[count-4];
          codes_nc[11] = nc_compress_codes[count-3];
          codes_nc[12] = nc_compress_codes[count-2];
          codes_nc[13] = nc_compress_codes[count-1];
          codes_nc[14] = nc_compress_codes[count-0];

          do_900_to_10();
          
        }         
     }  /* while codeindex < */


}

void do_main_loop()

{
int first_in;
 int debug;

 debug = 0;

 init_char_arrays();

 end_found = FALSE;

 codeindex = 0;

 code = get_next_code();

 number_of_codes = code;

 code = get_next_code();

 first_in = TRUE;

 if (debug) { printf("In main_loop : number_of_codes = %d \n", 
		     number_of_codes ); }

 while ( end_found == FALSE )
   {
   
     if (( code == 900 ) || ( first_in == TRUE))
       {
         do_text_compress ();
         first_in = FALSE;
       }
     if ( code == 901 )       // codes not a multiple of 6
       {
         do_byte_compress(901);   
         first_in = FALSE;
       }
     if ( code == 902 )
       {
        do_nc_compress();
        first_in = FALSE;
       }
     if ( code == 913 )      // temporary jump to byte compress
       {
         do_byte_compress(913);
         first_in = FALSE;
       }
     if ( code == 924 )     // codes a multiple of 6
       {
         do_byte_compress(924);
         first_in = FALSE;

       }      

     if ( code < 900)
       {
	 printf("Expected code at least = 900 \n");
         code = get_next_code();
       }     
   }
}

/*
 * Performs ERRORS+ERASURES decoding of RS codes. If decoding is successful,
 * writes the codeword into data[] itself. Otherwise data[] is unaltered.
 *
 * Return number of symbols corrected, or -1 if codeword is illegal
 * or uncorrectable. If eras_pos is non-null, the detected error locations
 * are written back. NOTE! This array must be at least NN-KK elements long.
 * 
 * First "no_eras" erasures are declared by the calling program. Then, the
 * maximum # of errors correctable is t_after_eras = floor((NN-KK-no_eras)/2).
 * If the number of channel errors is not greater than "t_after_eras" the
 * transmitted codeword will be recovered. Details of algorithm can be found
 * in R. Blahut's "Theory ... of Error-Correcting Codes".

 * Warning: the eras_pos[] array must not contain duplicate entries; decoder failure
 * will result. The decoder *could* check for this condition, but it would involve
 * extra time on every decoding operation.
 */

int eras_dec_rs(int *data, int *eras_pos, int no_eras, int data_len,
           int synd_len)
{
  int deg_lambda, el, deg_omega;
  int i, j, r,k;
  int u,q,tmp,num1,num2,den,discr_r;
  int lambda[2048 + 1], s[2048 + 1];	/* Err+Eras Locator poly
					 * and syndrome poly */
  int b[2048 + 1], t[2048 + 1], omega[2048 + 1];
  int root[2048], reg[2048 + 1], loc[2048];
  int syn_error, count;
  int ci;
  int mismatch;
  int error_val;
  int fix_loc; 
  int debug;

  debug = 0;
  

  /* Check for illegal input values */

  for(i=0;i< data_len;i++)
    {
    if(data[i] > GPRIME)
      {
	printf("Illegal data value = %d at location %d \n", data[i],i);
      }
    }


  /* form the syndromes; i.e., evaluate data(x) at roots of g(x)
   * namely @**(1+i)*PRIM, i = 0, ... ,(NN-KK-1)
   */
  for(i=1;i<=synd_len;i++){
    s[i] = data[data_len];
  }

  for(j=1;j<= data_len;j++){
    if(data[data_len -j] == 0)
      continue;
    tmp = Index_of[data[data_len-j]];
       
    /*	s[i] ^= Alpha_to[modbase(tmp + (1+i-1)*j)]; */
    for(i=1;i<=synd_len;i++)
      {
        s[i] = (s[i] + Alpha_to[modbase(tmp + (i)*j)]) % GPRIME;
       
      }
  }

  mismatch = FALSE;
  for ( j= 0; j < synd_len ; j += 1)
    {
      if ( s[j + 1] != synd_array[j])
        {
	  printf("Syndrome mismatch j = %d s[] = %d synd[] = %d \n",
		 j, s[j+1], synd_array[j]);
	  mismatch = TRUE;

        }
    }

  if ( mismatch == FALSE)
    {
      if (debug) { printf("Correct syndromes \n"); }
    }

  /* Convert syndromes to index form, checking for nonzero condition */
  syn_error = 0;
  for(i=1;i<=synd_len;i++){
    syn_error |= s[i];
    if (debug) { printf("Raw syndrome = %d i = %d \n", s[i],i ); }
    s[i] = Index_of[s[i]];
    
  }
  
  if (!syn_error) {
    /* if syndrome is zero, data[] is a codeword and there are no
     * errors to correct. So return data[] unmodified
     */
    count = 0;
    if ( verbose_out ) { printf("No errors \n"); }
    goto finish;
  }
  // CLEAR(&lambda[1],synd_len);

  for(ci=synd_len-1;ci >=0;ci--)
    lambda[ci+1] = 0;

  lambda[0] = 1;

  if (no_eras > 0) {
    /* Init lambda to be the erasure locator polynomial */
    lambda[1] = Alpha_to[modbase(PRIM * eras_pos[0])];
    for (i = 1; i < no_eras; i++) {
      u = modbase(PRIM*eras_pos[i]);
      for (j = i+1; j > 0; j--) {
	tmp = Index_of[lambda[j - 1]];
	if(tmp != A0)
	  lambda[j] = (lambda[j] + Alpha_to[modbase(u + tmp)]) % GPRIME;
      }
    }
#if DEBUG >= 1
    /* Test code that verifies the erasure locator polynomial just constructed
       Needed only for decoder debugging. */
    
    /* find roots of the erasure location polynomial */
    for(i=1;i<=no_eras;i++)
      reg[i] = Index_of[lambda[i]];
    count = 0;                       // usg NN = GPRIME-1?

    for (i = 1,k=data_len-Ldec; i <= data_len + synd_len; 
                         i++,k = modbase(data_len+k-Ldec))
     {
      q = 1;
      for (j = 1; j <= no_eras; j++)
	if (reg[j] != A0)
        {
	  reg[j] = modbase(reg[j] + j);
	  q = (q + Alpha_to[reg[j]]) % GPRIME;
	}
      if (q != 0)
	continue;
      /* store root and error location number indices */
      root[count] = i;
      loc[count] = k;
      count++;
    }
    if (count != no_eras) {
      printf("\n lambda(x) is WRONG\n");
      count = -1;
      goto finish;
    }
    if (debug) 
    {
     printf("\n Erasure positions as determined by roots of Eras Loc Poly:\n");
     for (i = 0; i < count; i++)
      {
       printf("loc  = %d ", loc[i]);
      }
     printf("\n");
    }
#endif
  }

  for(i=0;i<synd_len+1;i++)
    b[i] = Index_of[lambda[i]];
  
  /*
   * Begin Berlekamp-Massey algorithm to determine error+erasure
   * locator polynomial
   */
  r = no_eras;
  el = no_eras;
  while (++r <= synd_len) 
   {	/* r is the step number */
    /* Compute discrepancy at the r-th step in poly-form */
    discr_r = 0;
    for (i = 0; i < r; i++)
     {
      if ((lambda[i] != 0) && (s[r - i] != A0)) 
	{
	  if (debug) { printf("do add Index_of[lambda[]] = %d \n", 
 			      Index_of[lambda[i]]); }
	  if ( i % 2 == 1)
	    {
	discr_r = ( discr_r +
               Alpha_to[modbase((Index_of[lambda[i]] + s[r-i]))]) % GPRIME;
            }
          else
	    {
	discr_r = ( discr_r + GPRIME - 
               Alpha_to[modbase((Index_of[lambda[i]] + s[r-i]))]) % GPRIME;
            }
        if (debug) 
          { printf("In loop - discr = %d i = %d r = %d lambda[i] = %d s[r-i] = %d \n",
		   discr_r, i, r, lambda[i], s[r-i] ); }
        }
     }
    if (debug) { printf("r = %d Discrepency = %d \n", r, discr_r); }

    discr_r = Index_of[discr_r];	/* Index form */

    if (discr_r == A0) 
     {
      /* 2 lines below: B(x) <-- x*B(x) */
      //  COPYDOWN(&b[1],b,synd_len);
      //
       if (debug) { printf("Discrepency = A0 \n"); }
      for(ci=synd_len-1;ci >=0;ci--)
        {
         b[ci+1] = b[ci];
        }
      b[0] = A0;
     }
    else 
     {
      /* 7 lines below: T(x) <-- lambda(x) - discr_r*x*b(x) */
      /*  the T(x) will become the next lambda */

      t[0] = lambda[0];
      for (i = 0 ; i < synd_len; i++) {
        if (debug) { printf("i = %d b[i] = %d \n", i, b[i]); }
	if(b[i] != A0)
	  {
	     t[i+1] =  (lambda[i+1] +
	              Alpha_to[modbase(discr_r +  b[i])]) % GPRIME;
 
 	      if (debug) {
 printf("New t[i+1] = %d lambda[i+1] = %d b[i] = %d i = %d discr_r = %d\n",
	t[i+1], lambda[i+1], b[i], i ,discr_r); }
          }
	else
	  {
            t[i+1] = lambda[i+1];
          }
	if (debug) {  printf("i = %d t[i+1] = %d lambda[i+1] = %d \n", i, t[i+1],lambda[i+1]); }
      }
      el=0;
    if (2 * el <= r + no_eras - 1) 
      {
        if (debug) { printf("Reached the el stuff, inv  el = %d r = %d \n", el, r); }
	el = r + no_eras - el;
	/*
	 * 2 lines below: B(x) <-- inv(discr_r) *
	 * lambda(x)
	 */
	for (i = 0; i <= synd_len; i++)
	  {
	
               if ( lambda[i] == 0)
		 {
                   b[i] = A0;
                 }
               else
                 {
        
                   b[i] = modbase(Index_of[lambda[i]] - discr_r + GPRIME -1);
                   if (debug) {
		     printf("Inverting le  b[i] = %d i = %d \n", b[i], i); }
                 }
          }
      }
     else
      {
        if (debug) {
	  printf("Reached the el stuff, x mul,   el = %d r = %d \n", el, r); }
	/* 2 lines below: B(x) <-- x*B(x) */
	//	COPYDOWN(&b[1],b,synd_len);
        for(ci=synd_len-1;ci >=0;ci--)
	  {
           b[ci+1] = b[ci];
          }
	b[0] = A0;
      }
      //      COPY(lambda,t,synd_len+1);

      for(ci=synd_len+1-1;ci >=0;ci--)
	{
          lambda[ci] = t[ci];
          if (debug) { printf("ci = %d Lambda = %d \n", ci, t[ci]); }
        }
    }
  }

  /* Convert lambda to index form and compute deg(lambda(x)) */
  deg_lambda = 0;
  for(i=0;i<synd_len+1;i++){
    

    lambda[i] = Index_of[lambda[i]];

    if ( lambda[i] !=  A0)
      {
	deg_lambda = i;
      }

    if (debug) { printf("Lambda in index form = %d \n", lambda[i]); }
     
  }

  if (debug) {
    printf("Determination of deg_lambda = %d \n", deg_lambda); }

  /*
   * Find roots of the error+erasure locator polynomial by Chien
   * Search
   */

   for(ci=synd_len-1;ci >=0;ci--)
     reg[ci+1] = lambda[ci+1];

  count = 0;		/* Number of roots of lambda(x) */
  for (i = 1,k= data_len -1; i <= GPRIME   ; i++ )
	
   {
    q = 1;
    if (debug) {
	  printf("Reg[j] = %d q = %d i = %d \n", reg[j], q, i ); }
    for (j = deg_lambda; j > 0; j--)
      {
        
      if (reg[j] != A0)
        {
	  if ( debug ) {  printf("loop Reg[j] pre = %d \n", reg[j]); }
	reg[j] = modbase(reg[j] + j);
	//	q = modbase( q +  Alpha_to[reg[j]]);
        if ( deg_lambda != 1)
	  {
	    if ( j % 2 == 0)
	      {
       	       q = ( q + Alpha_to[reg[j]]) % GPRIME ;
             }
            else
              {
		q = ( q + GPRIME - Alpha_to[reg[j]])% GPRIME ;
              }
          }
        else
          {
            q = Alpha_to[reg[j]] % GPRIME;
            if ( q  == 1 )
              {
                q = q - 1;
              }
          }
        if (debug) {
	  printf("loop Reg[j] = %d q = %d i = %d j = %d %d = k\n", reg[j], q, i, j, k); }
        }
      }
    
    if (q == 0)
      {
    /* store root (index-form) and error location number */
        root[count] = i;

       loc[count] = GPRIME - 1 - i;
       if ( count < synd_len) 
	 {
	   count += 1;
         }
       else
         {
	   printf("Error : Error count too big = %d \n", count);
         }

       if (debug) {
         printf("root  = %d loc = %d \n", i, k ); }
  
      }
     if (k == 0)
      {
	k = data_len  - 1;
      }
    else
       {
        k -=1;
      }
    
    /* If we've already found max possible roots,
     * abort the search to save time
     */
        
     if(count == deg_lambda)
       {
	 break;
       }
  }
  if (deg_lambda != count) {
    /*
     * deg(lambda) unequal to number of roots => uncorrectable
     * error detected
     */
    
    printf("Uncorrectable error: root count = %d deg lambda = %d \n",
                         count,deg_lambda);
    count = -1;
    goto finish;
  }
  /*
   * Compute err+eras evaluator poly omega(x) = s(x)*lambda(x) (modulo
   * x**(synd_len)). in index form. Also find deg(omega).
   */
  deg_omega = 0;
  for (i = 0; i < synd_len; i++)
  {
    tmp = 0;
    j = (deg_lambda < i) ? deg_lambda : i;
    if (debug) {
    printf("j = %d deg_lambda = %d lambda[j] = %d \n",
	   j, deg_lambda, lambda[j]); }
    for(;j >= 0; j--){
          if ((s[i + 1 - j] != A0) && (lambda[j] != A0))
           {
	     if ( j % 2 == 1 )
               {
                  
	         tmp = ( tmp + GPRIME -
	            Alpha_to[modbase(s[i+1-j] + lambda[j])]) % GPRIME;
               }
             else
	       {
      
	     tmp = (tmp + Alpha_to[modbase(s[i+ 1 - j] + lambda[j])]) % GPRIME;
               }
	  if (debug) {
         printf("In tmp loop  tmp = %d i = %d j = %d s[i+1-j] = %d lambda[j] = %d \n",
		tmp, i, j, s[i+1-j], lambda[j] ); }
        }
    }
    if(tmp != 0)
      deg_omega = i;
    omega[i] = Index_of[tmp];
    if (debug) { printf("Omega [i] = %d i = %d \n", omega[i], i); }

  }
  omega[synd_len] = A0;
  if (debug) {
    printf("Degree of omega = %d \n", deg_omega); }
  
  /*
   * Compute error values in poly-form. num1 = omega(inv(X(l))), num2 =
   * inv(X(l))**(B0-1) and den = lambda_pr(inv(X(l))) all in poly-form
   */
  for (j = count-1; j >=0; j--)
   {
    num1 = 0;
    for (i = deg_omega; i >= 0; i--) 
      {
      if (omega[i] != A0)
        {
	  //	num1  = ( num1 + Alpha_to[modbase(omega[i] + (i * root[j])]) % GPRIME;
	num1  = ( num1 + Alpha_to[modbase(omega[i] + ((i+1) * root[j]))]) % GPRIME;
        if (debug) { printf("Num1 = %d i = %d omega[i] = %d root[j] = %d \n",
			    num1, i, omega[i], root[j] ); }
        }
      }
   //  num2 = Alpha_to[modbase(root[j] * (1 - 1) + data_len)];
   
    num2 = 1;
    den = 0;
    
    // denominator if product of all (1 - Bj Bk) for k != j
    // if count = 1, then den = 1

    den  = 1;
    for ( k = 0; k < count ; k += 1)
      {
	if ( k != j)
              {
                 tmp = ( 1 + GPRIME - Alpha_to[modbase(GPRIME -1 - root[k] + root[j])])%GPRIME;
               
                den = Alpha_to[ modbase( Index_of[den] + Index_of[tmp])];
	      }
      }

    if (debug) { printf("den = %d \n", den); }

    if (den == 0) {
      if (debug) {
	printf("\n ERROR: denominator = 0\n"); }

      /* Convert to dual- basis */
      count = -1;
      goto finish;
    }

    if (debug) { printf("Index num1 = %d Index num2 = %d Index of den = %d \n",
			Index_of[num1], Index_of[num2], Index_of[den]); }

    error_val = Alpha_to[modbase(Index_of[num1] + Index_of[num2] +
		GPRIME - 1 - Index_of[den])] % GPRIME;

    if (debug)  { printf("error_val = %d \n", error_val); }

    /* Apply error to data */
    if (num1 != 0) {
      if ( loc[j] < data_len + 1 )
	{
	  fix_loc = data_len   - loc[j];
          
          if (debug) { printf("Fix loc = %d \n", fix_loc); }
          if ( fix_loc < (data_len + 1))
	    {
             printf("Correcting data at %d , old data = %d ", fix_loc, 
		    data[fix_loc]);
             data[fix_loc] = (data[fix_loc] + GPRIME -  error_val)%GPRIME;
             printf("new data = %d \n", 
		    data[fix_loc]);
            }
        }
    }
  }
 finish:
  if (debug) { printf("At FINISH \n"); }
    if(eras_pos != NULL)
      {
       for(i=0;i<count;i++)
         {
           if(eras_pos!= NULL)
	     eras_pos[i] = loc[i];
         }
       }
    return count;
}

void do_scan()
{

  char txt_file_str[120];

  use_image = FALSE;

  verbose_out = 0;

  init_powers_of_900();

 //  generate the t_array

  convert_to_t();

  powers_init();   // needed for error correction 

  strcpy(txt_file_str,"out.txt");
          
 outfile1=fopen(txt_file_str,"w");

 if ( outfile1 == NULL)
   {
     printf("Unable to open the output file %s \n", txt_file_str);
     exit(0);
   }


  get_preprocessor_data();

  printf("Doing rows = %d columns = %d \n",
	rows_from_raw, columns_from_raw);

 do_rows_columns(rows_from_raw, columns_from_raw);


 do_main_loop();

 fclose( outfile1);

}

int main(int argc, char **argv)
{
  
  out_cnt = 0;
  bestfig = 1000;

  inargstr[0] = 0;

  if ( argc > 2)
    {
      strcpy(inargstr, argv[2]);

      if ( isdigit( inargstr[0] ))
	{
          strcpy(pgm_infile_str, argv[1]);
        }
      else
        {
	  printf("Error:  Argument two should be a number 1-7 not %s\n",
                    inargstr);
          exit(-1);
        }
    }
  else
    {
      if ( argc == 2)
	{
          strcpy(pgm_infile_str, argv[1]);
        }
    }

  // printf("inargstr = %s \n",inargstr);
  write_mods = 0;
  write_final_pix = 0;
  write_leftcol = TRUE;
  write_cols = 0;
  write_outpix = 0;
  write_data;


  // inargstr 0 = no print
  //          1 bit set  = write all the mods
  //          2 bit set = write final pix
  //          4 bit set = write output pix and all column pix
  if ( strcmp(inargstr,"1") == 0)
    {
      write_mods = 1;
    }

  if ( strcmp(inargstr,"2") == 0)
    {
      write_final_pix = 1;
    }

  if ( strcmp(inargstr,"3") == 0)
    {
      write_final_pix = 1;
      write_mods = 1;
    }

  if ( strcmp(inargstr,"4") == 0)
    {
      write_outpix = 1;
      write_cols = 1;
    }

  if ( strcmp(inargstr,"5") == 0)
    {
      write_mods = 1;
      write_outpix = 1;
      write_cols = 1;
    }

  if ( strcmp(inargstr,"6") == 0)
    {
      write_final_pix = 1;
      write_outpix = 1;
      write_cols = 1;
    }

  if ( strcmp(inargstr,"7") == 0)
    {
      write_final_pix = 1;
      write_mods = 1;
      write_outpix = 1;
      write_cols = 1;
    }

  if ( strcmp(inargstr,"8") == 0)
    {
      write_final_pix = 1;
      write_mods = 1;
      write_outpix = 1;
      write_cols = 1;
      write_data = 1;
    }

  line_threshold = 60;

  fail = read_pgm_in();

  //  pgm_init( &argc, argv);
  
  if ( fail == 0)
    {
      exit(0);
    }

  do_prelim();

  // do what was done by scan program

  if ( pb_index > 0)
    {
     do_scan(); 
    }
  else
    {
      printf("Unable to decode bar code \n");
    }
}
